# Form implementation generated from reading ui file 'main_dialog_base.ui'
#
# Created by: PyQt6 UI code generator 6.4.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PySide6 import QtCore, QtGui, QtWidgets


class Ui_MainDialogBase(object):
  def setupUi(self, MainDialogBase):
    MainDialogBase.setObjectName("MainDialogBase")
    MainDialogBase.resize(1261, 782)
    self.gridLayout = QtWidgets.QGridLayout(MainDialogBase)
    self.gridLayout.setObjectName("gridLayout")
    self.buttonBox = QtWidgets.QDialogButtonBox(parent=MainDialogBase)
    self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
    self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel)
    self.buttonBox.setObjectName("buttonBox")
    self.gridLayout.addWidget(self.buttonBox, 3, 0, 1, 1)
    self.mainHorizontalLayout = QtWidgets.QHBoxLayout()
    self.mainHorizontalLayout.setObjectName("mainHorizontalLayout")
    self.mainVerticalLayout = QtWidgets.QVBoxLayout()
    self.mainVerticalLayout.setObjectName("mainVerticalLayout")
    self.scrollAreaHorizontalLayout = QtWidgets.QHBoxLayout()
    self.scrollAreaHorizontalLayout.setObjectName("scrollAreaHorizontalLayout")
    self.projectsScrollArea = QtWidgets.QScrollArea(parent=MainDialogBase)
    self.projectsScrollArea.setWidgetResizable(True)
    self.projectsScrollArea.setObjectName("projectsScrollArea")
    self.projectScrollAreaWidgetContents = QtWidgets.QWidget()
    self.projectScrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 1235, 336))
    self.projectScrollAreaWidgetContents.setObjectName("projectScrollAreaWidgetContents")
    self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.projectScrollAreaWidgetContents)
    self.verticalLayout_3.setObjectName("verticalLayout_3")
    self.projectsScrollAreaVerticalLayout = QtWidgets.QVBoxLayout()
    self.projectsScrollAreaVerticalLayout.setObjectName("projectsScrollAreaVerticalLayout")
    self.verticalLayout_3.addLayout(self.projectsScrollAreaVerticalLayout)
    self.projectsScrollArea.setWidget(self.projectScrollAreaWidgetContents)
    self.scrollAreaHorizontalLayout.addWidget(self.projectsScrollArea)
    self.mainVerticalLayout.addLayout(self.scrollAreaHorizontalLayout)
    self.uploadButtonHorizontalLayout = QtWidgets.QHBoxLayout()
    self.uploadButtonHorizontalLayout.setObjectName("uploadButtonHorizontalLayout")
    spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
    self.uploadButtonHorizontalLayout.addItem(spacerItem)
    self.selectAllPushButton = QtWidgets.QPushButton(parent=MainDialogBase)
    self.selectAllPushButton.setMinimumSize(QtCore.QSize(100, 0))
    self.selectAllPushButton.setObjectName("selectAllPushButton")
    self.uploadButtonHorizontalLayout.addWidget(self.selectAllPushButton)
    self.deselectAllPushButton = QtWidgets.QPushButton(parent=MainDialogBase)
    self.deselectAllPushButton.setMinimumSize(QtCore.QSize(100, 0))
    self.deselectAllPushButton.setObjectName("deselectAllPushButton")
    self.uploadButtonHorizontalLayout.addWidget(self.deselectAllPushButton)
    self.configureUploadPushButton = QtWidgets.QPushButton(parent=MainDialogBase)
    self.configureUploadPushButton.setMinimumSize(QtCore.QSize(100, 0))
    self.configureUploadPushButton.setObjectName("configureUploadPushButton")
    self.uploadButtonHorizontalLayout.addWidget(self.configureUploadPushButton)
    self.editFullMetadataPushButton = QtWidgets.QPushButton(parent=MainDialogBase)
    self.editFullMetadataPushButton.setMinimumSize(QtCore.QSize(100, 0))
    self.editFullMetadataPushButton.setObjectName("editFullMetadataPushButton")
    self.uploadButtonHorizontalLayout.addWidget(self.editFullMetadataPushButton)
    self.uploadPushButton = QtWidgets.QPushButton(parent=MainDialogBase)
    self.uploadPushButton.setMinimumSize(QtCore.QSize(100, 0))
    self.uploadPushButton.setObjectName("uploadPushButton")
    self.uploadButtonHorizontalLayout.addWidget(self.uploadPushButton)
    self.mainVerticalLayout.addLayout(self.uploadButtonHorizontalLayout)
    self.scrollAreaVerticalLayout = QtWidgets.QVBoxLayout()
    self.scrollAreaVerticalLayout.setObjectName("scrollAreaVerticalLayout")
    self.uploadQueueScrollArea = QtWidgets.QScrollArea(parent=MainDialogBase)
    self.uploadQueueScrollArea.setWidgetResizable(True)
    self.uploadQueueScrollArea.setObjectName("uploadQueueScrollArea")
    self.scrollAreaWidgetContents = QtWidgets.QWidget()
    self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 1235, 302))
    self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
    self.verticalLayout = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
    self.verticalLayout.setObjectName("verticalLayout")
    self.uploadQueueVerticalLayout = QtWidgets.QVBoxLayout()
    self.uploadQueueVerticalLayout.setObjectName("uploadQueueVerticalLayout")
    self.verticalLayout.addLayout(self.uploadQueueVerticalLayout)
    self.uploadQueueScrollArea.setWidget(self.scrollAreaWidgetContents)
    self.scrollAreaVerticalLayout.addWidget(self.uploadQueueScrollArea)
    self.scrollAreaButtonsHorizontalLayout = QtWidgets.QHBoxLayout()
    self.scrollAreaButtonsHorizontalLayout.setObjectName("scrollAreaButtonsHorizontalLayout")
    spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
    self.scrollAreaButtonsHorizontalLayout.addItem(spacerItem1)
    self.clearFinishedPushButton = QtWidgets.QPushButton(parent=MainDialogBase)
    self.clearFinishedPushButton.setMinimumSize(QtCore.QSize(150, 0))
    self.clearFinishedPushButton.setObjectName("clearFinishedPushButton")
    self.scrollAreaButtonsHorizontalLayout.addWidget(self.clearFinishedPushButton)
    self.cancelAllPushButton = QtWidgets.QPushButton(parent=MainDialogBase)
    self.cancelAllPushButton.setMinimumSize(QtCore.QSize(150, 0))
    self.cancelAllPushButton.setObjectName("cancelAllPushButton")
    self.scrollAreaButtonsHorizontalLayout.addWidget(self.cancelAllPushButton)
    self.showCompletedPushButton = QtWidgets.QPushButton(parent=MainDialogBase)
    self.showCompletedPushButton.setMinimumSize(QtCore.QSize(150, 0))
    self.showCompletedPushButton.setObjectName("showCompletedPushButton")
    self.scrollAreaButtonsHorizontalLayout.addWidget(self.showCompletedPushButton)
    self.scrollAreaVerticalLayout.addLayout(self.scrollAreaButtonsHorizontalLayout)
    self.mainVerticalLayout.addLayout(self.scrollAreaVerticalLayout)
    self.mainHorizontalLayout.addLayout(self.mainVerticalLayout)
    self.gridLayout.addLayout(self.mainHorizontalLayout, 0, 0, 1, 1)
    spacerItem2 = QtWidgets.QSpacerItem(20, 5, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Fixed)
    self.gridLayout.addItem(spacerItem2, 1, 0, 1, 1)

    self.retranslateUi(MainDialogBase)
    self.buttonBox.accepted.connect(MainDialogBase.accept) # type: ignore
    self.buttonBox.rejected.connect(MainDialogBase.reject) # type: ignore
    QtCore.QMetaObject.connectSlotsByName(MainDialogBase)

  def retranslateUi(self, MainDialogBase):
    _translate = QtCore.QCoreApplication.translate
    MainDialogBase.setWindowTitle(_translate("MainDialogBase", "Dataverse upload"))
    self.selectAllPushButton.setToolTip(_translate("MainDialogBase", "Select all the projects listed above."))
    self.selectAllPushButton.setText(_translate("MainDialogBase", "Select all"))
    self.deselectAllPushButton.setToolTip(_translate("MainDialogBase", "Deselect all the projects listed above."))
    self.deselectAllPushButton.setText(_translate("MainDialogBase", "Deselect all"))
    self.configureUploadPushButton.setToolTip(_translate("MainDialogBase", "Configure the contents of projects which need to be uploaded to dataverse."))
    self.configureUploadPushButton.setText(_translate("MainDialogBase", "Configure"))
    self.editFullMetadataPushButton.setToolTip(_translate("MainDialogBase", "Click to edit full/minimal list of metadata to be used for dataverse upload."))
    self.editFullMetadataPushButton.setText(_translate("MainDialogBase", "Edit metadata"))
    self.uploadPushButton.setToolTip(_translate("MainDialogBase", "Start uploading the above selected projects to dataverse."))
    self.uploadPushButton.setText(_translate("MainDialogBase", "Start upload"))
    self.uploadQueueScrollArea.setToolTip(_translate("MainDialogBase", "<html><head/><body><p><span style=\" font-style:italic;\">Displays the enqueued lists of PASTA projects to be uploaded to dataverse.</span><span style=\" font-style:italic;\">Users can view individual logs for each project upload, cancel each or all of them and also clear all the finished items anytime.</span></p></body></html>"))
    self.clearFinishedPushButton.setToolTip(_translate("MainDialogBase", "Clear all finished/cancelled/error uploads."))
    self.clearFinishedPushButton.setText(_translate("MainDialogBase", "Clear"))
    self.cancelAllPushButton.setToolTip(_translate("MainDialogBase", "Cancel all the ongoing uploads."))
    self.cancelAllPushButton.setText(_translate("MainDialogBase", "Cancel all"))
    self.showCompletedPushButton.setToolTip(_translate("MainDialogBase", "Show uploaded projects history."))
    self.showCompletedPushButton.setText(_translate("MainDialogBase", "Show completed"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    MainDialogBase = QtWidgets.QDialog()
    ui = Ui_MainDialogBase()
    ui.setupUi(MainDialogBase)
    MainDialogBase.show()
    sys.exit(app.exec())



# pylint: skip-file
# mypy: ignore-errors
import hashlib
import io
from typing import (
    BinaryIO,
    Union,
)

from .exceptions import (
    Error,
    ParseError,
)


class Reader:
    def __init__(self, data: bytes):
        self._buf = data
        self._pos = 0

    def __len__(self) -> int:
        return len(self._buf) - self._pos

    def read(self, size: int) -> bytes:
        pos = self._pos + size
        data = self._buf[self._pos:pos]
        if len(data) != size:
            raise ParseError('read size mismatch')
        self._pos = pos
        return data


def read_data(data: Union[bytes, BinaryIO], prehash: bool) -> bytes:
    if prehash:
        if isinstance(data, io.BufferedIOBase):
            hasher = hashlib.blake2b()
            while chunk := data.read(1 << 13):
                hasher.update(chunk)
            data = hasher.digest()
        else:
            data = hashlib.blake2b(data).digest()
    elif isinstance(data, io.BufferedIOBase):
        data = data.read()
    return data


def check_comment(s: str):
    if '\n' in s:
        raise Error('comment contains new line char')



""" Creates a new instance of the EditMetadataSummaryDialog class. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: edit_metadata_summary_dialog.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
import logging
from typing import Any, Callable

from PySide6 import QtCore
from PySide6.QtWidgets import QDialog

from pasta_eln.GUI.dataverse.edit_metadata_summary_dialog_base import Ui_EditMetadataSummaryDialog


class EditMetadataSummaryDialog(Ui_EditMetadataSummaryDialog):
  """
  Creates a new instance of the EditMetadataSummaryDialog class.
  """

  def __new__(cls, *_: Any, **__: Any) -> Any:
    """
    Creates a new instance of the EditMetadataSummaryDialog class.

    Explanation:
        This method creates a new instance of the EditMetadataSummaryDialog class.

    Args:
        *_: Variable length argument list.
        **__: Arbitrary keyword arguments.

    Returns:
        Any: The new instance of the EditMetadataDialog class.

    """
    return super(EditMetadataSummaryDialog, cls).__new__(cls)

  def __init__(self, save_config_callback: Callable[[], None]) -> None:
    """
    Initializes the EditMetadataSummaryDialog instance.

    Args:
        save_config_callback (Callable[[], None]): A callable function to be connected to the 'accepted' signal of the buttonBox.

    Raises:
        TypeError: If save_config_callback is not callable.
    """
    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
    if not callable(save_config_callback):
      raise TypeError("save_config_callback must be callable.")
    self.instance = QDialog()
    super().setupUi(self.instance)
    self.instance.setWindowModality(QtCore.Qt.WindowModality.ApplicationModal)
    self.summaryTextEdit.setReadOnly(True)
    self.buttonBox.accepted.connect(save_config_callback)

  def show(self) -> None:
    """
    Displays the EditMetadataSummaryDialog instance.
    """
    self.instance.show()



""" New/Edit dialog (dialog is blocking the main-window, as opposed to create a new widget-window)"""
import logging, re, copy, json
from enum import Enum
from pathlib import Path
from typing import Any, Union
from PySide6.QtWidgets import QDialog, QWidget, QVBoxLayout, QHBoxLayout, QScrollArea, QSplitter  # pylint: disable=no-name-in-module
from PySide6.QtWidgets import QLabel, QTextEdit, QTabWidget, QPlainTextEdit, QComboBox, QLineEdit     # pylint: disable=no-name-in-module
from PySide6.QtWidgets import QSizePolicy, QMessageBox# pylint: disable=no-name-in-module
from PySide6.QtGui import QRegularExpressionValidator # pylint: disable=no-name-in-module
from PySide6.QtCore import QSize, Qt, QTimer          # pylint: disable=no-name-in-module
from ..guiStyle import Image, TextButton, IconButton, Label, showMessage, widgetAndLayout, widgetAndLayoutForm, ScrollMessageBox
from ._contextMenu import initContextMenu, executeContextMenu, CommandMenu
from ..fixedStringsJson import defaultDataHierarchyNode
from ..miscTools import createDirName, markdownStyler
from ..guiCommunicate import Communicate

class Form(QDialog):
  """ New/Edit dialog (dialog is blocking the main-window, as opposed to create a new widget-window)"""
  def __init__(self, comm:Communicate, doc:dict[str,Any]):
    """
    Initialization

    Args:
      comm (Communicate): communication channel
      doc (dict):  document to change / create
    """
    super().__init__()
    self.comm = comm
    self.db   = self.comm.backend.db
    self.doc  = copy.deepcopy(doc)
    if '_attachments' in self.doc:
      del self.doc['_attachments']
    self.flagNewDoc = True
    if '_id' in self.doc or '_ids' in self.doc:
      self.flagNewDoc = False
    if self.flagNewDoc:
      self.setWindowTitle('Create new entry')
      self.doc['-name'] = ''
    else:
      self.setWindowTitle('Edit information')
    self.skipKeys = ['image','metaVendor','metaUser','shasum']
    self.skipKeys0= ['_','-', '#']
    self.allHidden = False

    # GUI elements
    mainL = QVBoxLayout(self)
    splitter = QSplitter(Qt.Orientation.Horizontal)
    splitter.setHandleWidth(10)
    splitter.setContentsMargins(0,0,0,0)
    mainL.addWidget(splitter, stretch=2)

    # image
    if 'image' in self.doc:
      imageWSA = QScrollArea()
      imageWSA.setWidgetResizable(True)
      imageW, self.imageL = widgetAndLayout('V', splitter)
      width = self.comm.backend.configuration['GUI']['imageSizeDetails'] \
                if hasattr(self.comm.backend, 'configuration') else 300
      Image(self.doc['image'], self.imageL, anyDimension=width)
      if '_id' in self.doc:
        self.docID= doc['_id']  #required for hide to work
        imageW.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        imageW.customContextMenuRequested.connect(lambda pos: initContextMenu(self, pos))
      imageWSA.setWidget(imageW)
      splitter.addWidget(imageWSA)
      self.setMinimumWidth(1000)
    else:
      self.setMinimumWidth(600)

    # create full data set
    if self.doc['-type'][0] in self.db.dataHierarchy:
      dataHierarchyNode = copy.deepcopy(self.db.dataHierarchy[self.doc['-type'][0]]['meta'])
    else:
      dataHierarchyNode = copy.deepcopy(defaultDataHierarchyNode)
    keysDataHierarchy = [i['name'] for group in dataHierarchyNode for i in dataHierarchyNode[group]]
    for keyInDocNotHierarchy in set(self.doc.keys()).difference(keysDataHierarchy ):
      dataHierarchyNode['default'].append({'name':keyInDocNotHierarchy})
    self.allKeys = {i['name'] for group in dataHierarchyNode for i in dataHierarchyNode[group]}
    self.allKeys = self.allKeys.union(self.doc.keys())

    # create tabs or not: depending on the number of groups
    if '-tags' not in self.doc:
      self.doc['-tags'] = []
    self.tabW = QTabWidget() #has count=0 if not connected
    if len(dataHierarchyNode)>1:
      self.tabW.setParent(self)
      self.tabW.tabBarClicked.connect(self.changeTabs)
      splitter.addWidget(self.tabW)

    # create forms by looping
    self.formsL = []
    for group in dataHierarchyNode:
      if len(dataHierarchyNode)==1:
        _, formL = widgetAndLayoutForm(splitter, 's')
      else:
        formW, formL = widgetAndLayoutForm(None, 's', top='m')
        self.tabW.addTab(formW, group if group!='default' else 'Home')
      self.formsL.append(formL)
      for key in [i['name'] for i in dataHierarchyNode[group]]:
        value = self.doc.get(key, '')

        # case list
        if key == '-name' and '_ids' not in self.doc:
          setattr(self, 'key_-name', QLineEdit(self.doc['-name']))
          getattr(self, 'key_-name').setValidator(QRegularExpressionValidator("[\\w\\ .-]+"))
          formL.addRow('Name', getattr(self, 'key_-name'))
        elif key == '-tags':
          self.tagsBarMainW, tagsBarMainL = widgetAndLayout('H', spacing='s')
          _, self.tagsBarSubL = widgetAndLayout('H', tagsBarMainL, spacing='s', right='m') #part which shows all the tags
          self.otherChoices = QComboBox()   #part/combobox that allow user to select
          self.otherChoices.setEditable(True)
          self.otherChoices.setMaximumWidth(100)
          self.otherChoices.setValidator(QRegularExpressionValidator("[a-z]\\w+"))
          self.otherChoices.setIconSize(QSize(0,0))
          self.otherChoices.setInsertPolicy(QComboBox.InsertPolicy.InsertAtBottom)
          tagsBarMainL.addWidget(self.otherChoices)
          self.gradeChoices = QComboBox()   #part/combobox that shows grades
          self.gradeChoices.setMaximumWidth(80)
          self.gradeChoices.setIconSize(QSize(0,0))
          self.gradeChoices.addItems(['','\u2605','\u2605'*2,'\u2605'*3,'\u2605'*4,'\u2605'*5])
          self.gradeChoices.currentTextChanged.connect(self.addTag)
          tagsBarMainL.addWidget(self.gradeChoices)
          formL.addRow(QLabel('Tags:'), self.tagsBarMainW)
          self.updateTagsBar()
          self.otherChoices.currentIndexChanged.connect(self.addTag) #connect to slot only after all painting is done
        elif key in ['comment','content']:
          labelW, labelL = widgetAndLayout('V')
          labelL.addWidget(QLabel(key.capitalize()))
          TextButton('More', self, [Command.FOCUS_AREA, key], labelL, checkable=True)
          rightSideW, rightSideL = widgetAndLayout('V')
          setattr(self, f'buttonBarW_{key}', QWidget())
          getattr(self, f'buttonBarW_{key}').hide()
          buttonBarL = QHBoxLayout(getattr(self, f'buttonBarW_{key}'))
          for name, tooltip in [['bold','Bold text'],['italic','Italic text'],['list-ul','Bullet list'],\
                                      ['list-ol','Numbered list']]:
            IconButton(f'fa5s.{name}', self, [Command.BUTTON_BAR, name, key], buttonBarL, tooltip)
          for i in range(1,4):
            icon = f'mdi.format-header-{str(i)}'
            IconButton(icon, self, [Command.BUTTON_BAR, f'heading{str(i)}', key], buttonBarL, f'Heading {str(i)}')
          rightSideL.addWidget(getattr(self, f'buttonBarW_{key}'))
          setattr(self, f'textEdit_{key}', QPlainTextEdit(value))
          getattr(self, f'textEdit_{key}').setAccessibleName(key)
          getattr(self, f'textEdit_{key}').setTabStopDistance(20)
          getattr(self, f'textEdit_{key}').textChanged.connect(self.textChanged)
          setattr(self, f'textShow_{key}', QTextEdit())
          getattr(self, f'textShow_{key}').setMarkdown(markdownStyler(value))
          getattr(self, f'textShow_{key}').setReadOnly(True)
          getattr(self, f'textShow_{key}').hide()
          splitter= QSplitter()
          splitter.setSizePolicy(QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding))
          splitter.addWidget(getattr(self, f'textEdit_{key}'))
          splitter.addWidget(getattr(self, f'textShow_{key}'))
          rightSideL.addWidget(splitter)
          formL.addRow(labelW, rightSideW)
        elif (key[0] in self.skipKeys0) or key in self.skipKeys:  #skip non desired ones
          continue
        elif isinstance(value, list):   #list of items, qrCodes in sample
          if len(value)>0 and isinstance(value[0], str):
            setattr(self, f'key_{key}', QLineEdit(' '.join(value)))
            formL.addRow(QLabel(key.capitalize()), getattr(self, f'key_{key}'))
          else:
            logging.info('Cannot display value of key=%s: %s. Write unknown value for docID=%s',
                         key, str(value), self.doc['_id'])
        elif isinstance(value, str):    #string
          dataHierarchyItem = [i for group in dataHierarchyNode for i in dataHierarchyNode[group] if i['name']==key]
          if len(dataHierarchyItem)==1 and 'list' in dataHierarchyItem[0]:       #choice dropdown
            setattr(self, f'key_{key}', QComboBox())
            if isinstance(dataHierarchyItem[0]['list'], list):            #dataHierarchy-defined choices
              getattr(self, f'key_{key}').addItems(dataHierarchyItem[0]['list'])
            else:                                                    #choice among docType
              listDocType = dataHierarchyItem[0]['list']
              getattr(self, f'key_{key}').addItem('- no link -', userData='')
              for line in self.db.getView(f'viewDocType/{listDocType}'):
                getattr(self, f'key_{key}').addItem(line['value'][0], userData=line['id'])
                if line['id'] == value:
                  getattr(self, f'key_{key}').setCurrentIndex(getattr(self, f'key_{key}').count()-1)
          else:                                   #text area
            setattr(self, f'key_{key}', QLineEdit(value))
          formL.addRow(QLabel(key.capitalize()), getattr(self, f'key_{key}'))
        else:
          print(f"**WARNING dialogForm: unknown value type. key:{key}, type:{type(value)}")
      if group == 'default':
        # individual key-value items
        self.keyValueListW, self.keyValueListL = widgetAndLayoutForm(None, 's')
        self.keyValueListW.hide()
        self.keyValueLabel = QLabel('Key - values')
        self.keyValueLabel.hide()
        formL.addRow(self.keyValueLabel, self.keyValueListW)
        # add extra questions at bottom of form
        allowProjectAndDocTypeChange = '_id' in self.doc and self.doc['-type'][0][0]!='x'
        if '_ids' in self.doc: #if group edit
          allowProjectAndDocTypeChange = all(docID[0] != 'x' for docID in self.doc['_ids'])
        if allowProjectAndDocTypeChange: #if not-new and non-folder
          formL.addRow(QLabel('Special properties:'), QLabel('') )
        label = '- unassigned -' if self.flagNewDoc else '- no change -'
        if allowProjectAndDocTypeChange or ('_id' not in self.doc and self.doc['-type'][0][0]!='x'): #if new and non-folder
          self.projectComboBox = QComboBox()
          self.projectComboBox.addItem(label, userData='')
          for line in self.db.getView('viewDocType/x0'):
            # add all projects but the one that is present
            if '-branch' not in self.doc or all( not(len(branch['stack'])>0 and line['id']==branch['stack'][0])
                                                for branch in self.doc['-branch']):
              self.projectComboBox.addItem(line['value'][0], userData=line['id'])
              if self.doc.get('_projectID','') == line['id']:
                self.projectComboBox.setCurrentIndex(self.projectComboBox.count()-1)
          formL.addRow(QLabel('Project'), self.projectComboBox)
          if '_projectID' in self.doc:
            del self.doc['_projectID']
        if allowProjectAndDocTypeChange: #if not-new and non-folder
          self.docTypeComboBox = QComboBox()
          self.docTypeComboBox.addItem(label, userData='')
          for key, value in self.db.dataLabels.items():
            if key[0]!='x':
              self.docTypeComboBox.addItem(value, userData=key)
          self.docTypeComboBox.addItem('_UNIDENTIFIED_', userData='-')
          formL.addRow(QLabel('Data type'), self.docTypeComboBox)
    if [i for i in self.doc.keys() if i.startswith('_')]:
      logging.error('There should not be "_" in a doc: '+str(self.doc))
    # final button box
    _, buttonLineL = widgetAndLayout('H', mainL, 'm')
    if '-branch' in self.doc:
      visibilityIcon = all(all(branch['show']) for branch in self.doc['-branch'])
      self.visibilityText = QLabel('' if visibilityIcon else 'HIDDEN     \U0001F441')
      buttonLineL.addWidget(self.visibilityText)
    buttonLineL.addStretch(1)
    self.btnAddKWPairs = IconButton('ri.menu-add-fill', self, [Command.FORM_ADD_KV],   buttonLineL,
                                    'Add key-value pair', style='border-width:1')
    IconButton('fa5s.poll-h',      self, [Command.FORM_SHOW_DOC], buttonLineL, 'Show all information',
               style='border-width:1')
    TextButton('Save',             self, [Command.FORM_SAVE],     buttonLineL, 'Save changes')
    TextButton('Cancel',           self, [Command.FORM_CANCEL],   buttonLineL, 'Discard changes')
    if self.flagNewDoc: #new dataset
      TextButton('Save && Next', self, [Command.FORM_SAVE_NEXT], buttonLineL, 'Save this and handle next')
    # end of creating form autosave
    if (Path.home()/'.pastaELN.temp').is_file():
      with open(Path.home()/'.pastaELN.temp', 'r', encoding='utf-8') as fTemp:
        content = json.loads(fTemp.read())
        if self.doc.get('_id', '') in content:
          ret = QMessageBox.information(self, 'Information', 'There is unsaved information from a prematurely '+
                    'closed form. Do you want to restore it?\n If you decline, the unsaved information will be'+
                    ' removed.',
                  QMessageBox.StandardButton.No | QMessageBox.StandardButton.Yes,
                  QMessageBox.StandardButton.Yes)
          if ret==QMessageBox.StandardButton.Yes:
            subContent = content[self.doc.get('_id', '')]
            for key in subContent.keys():
              if key in ('comment', 'content'):
                getattr(self, f'textEdit_{key}').setPlainText(subContent[key])
              elif key in ('-tags'):
                self.doc[key] = subContent[key]
                self.updateTagsBar()
              elif isinstance(getattr(self, f'key_{key}'), QLineEdit):
                getattr(self, f'key_{key}').setText(subContent[key])
              # skip QCombobox items since cannot be sure that next from has them and they are easy to recreate
          del content[self.doc.get('_id', '')]
      with open(Path.home()/'.pastaELN.temp', 'w', encoding='utf-8') as fTemp:
        fTemp.write(json.dumps(content))
    self.checkThreadTimer = QTimer(self)
    self.checkThreadTimer.setInterval(1*60*1000) #1 min
    self.checkThreadTimer.timeout.connect(self.autosave)
    self.checkThreadTimer.start()


  def autosave(self) -> None:
    """ Autosave comment to file """
    if self.comm.backend.configuration['GUI']['autosave'] == 'No':
      return
    subContent = {'-name':getattr(self, 'key_-name').text().strip(), '-tags':self.doc['-tags']}
    for key in self.allKeys:
      if key in ['comment','content']:
        subContent[key] = getattr(self, f'textEdit_{key}').toPlainText().strip()
      elif key[0] in self.skipKeys0 or key in self.skipKeys or not hasattr(self, f'key_{key}'):
        continue
      elif isinstance(getattr(self, f'key_{key}'), QLineEdit):
        subContent[key] = getattr(self, f'key_{key}').text().strip()
      # skip QCombobox items since cannot be sure that next from has them and they are easy to recreate
    if (Path.home()/'.pastaELN.temp').is_file():
      with open(Path.home()/'.pastaELN.temp', 'r', encoding='utf-8') as fTemp:
        content = json.loads(fTemp.read())
    else:
      content = {}
    content[self.doc.get('_id', '')] = subContent
    with open(Path.home()/'.pastaELN.temp', 'w', encoding='utf-8') as fTemp:
      fTemp.write(json.dumps(content))
    return


  def execute(self, command:list[Any]) -> None:
    """
    Event if user clicks button

    Args:
      command (list): list of commands
    """
    if isinstance(command[0], CommandMenu):
      if executeContextMenu(self, command):
        self.imageL.itemAt(0).widget().setParent(None)
        width = self.comm.backend.configuration['GUI']['imageSizeDetails'] \
                if hasattr(self.comm.backend, 'configuration') else 300
        Image(self.doc['image'], self.imageL, anyDimension=width)
        visibilityIcon = all(all(branch['show']) for branch in self.doc['-branch'])
        self.visibilityText.setText('' if visibilityIcon else 'HIDDEN     \U0001F441')
    elif command[0] is Command.BUTTON_BAR:
      if command[1]=='bold':
        getattr(self, f'textEdit_{command[2]}').insertPlainText('**TEXT**')
      elif command[1]=='italic':
        getattr(self, f'textEdit_{command[2]}').insertPlainText('*TEXT*')
      elif command[1]=='list-ul':
        getattr(self, f'textEdit_{command[2]}').insertPlainText('\n- item 1\n- item 2')
      elif command[1]=='list-ol':
        getattr(self, f'textEdit_{command[2]}').insertPlainText('\n1. item 1\n1. item 2')
      elif command[1].startswith('heading'):
        getattr(self, f'textEdit_{command[2]}').insertPlainText('#' * int(command[1][-1]) +' Heading\n')
    elif command[0] is Command.FOCUS_AREA:
      unknownWidget = []
      idx = 0 if self.tabW.count()==0 else self.tabW.currentIndex()
      if self.allHidden:  #hide the special buttons and show general form
        getattr(self, f'textShow_{command[1]}').hide()
        getattr(self, f'buttonBarW_{command[1]}').hide()
        for i in range(self.formsL[idx].count()):
          widget = self.formsL[idx].itemAt(i).widget()
          if isinstance(widget, (QLabel, QComboBox, QLineEdit)):
            widget.show()
          else:
            unknownWidget.append(i)
        self.tagsBarMainW.show()
        if command[1]=='content' and len(unknownWidget)==5:  #show / hide label and right-side of non-content and non-comment
          self.formsL[idx].itemAt(unknownWidget[0]).widget().show()
          self.formsL[idx].itemAt(unknownWidget[1]).widget().show()
        if command[1]=='comment' and len(unknownWidget)==5:
          self.formsL[idx].itemAt(unknownWidget[2]).widget().show()
          self.formsL[idx].itemAt(unknownWidget[3]).widget().show()
        if self.keyValueListL.count() == 0:
          self.keyValueLabel.hide()
          self.keyValueListW.hide()
      else:  #show buttons to allow for easy markdown edit; hide general form
        getattr(self, f'textShow_{command[1]}').show()
        getattr(self, f'buttonBarW_{command[1]}').show()
        for i in range(self.formsL[idx].count()):
          widget = self.formsL[idx].itemAt(i).widget()
          if isinstance(widget, (QLabel, QComboBox, QLineEdit)):
            widget.hide()
          else:
            unknownWidget.append(i)
        self.tagsBarMainW.hide()
        if command[1]=='content' and len(unknownWidget)==5:
          self.formsL[idx].itemAt(unknownWidget[0]).widget().hide()
          self.formsL[idx].itemAt(unknownWidget[1]).widget().hide()
        if command[1]=='comment' and len(unknownWidget)==5:
          self.formsL[idx].itemAt(unknownWidget[2]).widget().hide()
          self.formsL[idx].itemAt(unknownWidget[3]).widget().hide()
      self.allHidden = not self.allHidden
    elif command[0] is Command.FORM_CANCEL:
      if self.comm.backend.configuration['GUI']['autosave'] == 'Yes':
        ret = QMessageBox.critical(self, 'Warning', 'You will lose the entered information. Do you want to '+
          'save everything to a temporary location?',
          QMessageBox.StandardButton.Cancel | QMessageBox.StandardButton.No | QMessageBox.StandardButton.Yes,
          QMessageBox.StandardButton.No)
        if ret==QMessageBox.StandardButton.Cancel:
          return
        if ret==QMessageBox.StandardButton.Yes:
          self.autosave()
      self.checkThreadTimer.stop()
      self.reject()
    elif command[0] in (Command.FORM_SAVE, Command.FORM_SAVE_NEXT):
      # create the data that has to be saved
      self.checkThreadTimer.stop()
      if (Path.home()/'.pastaELN.temp').is_file():
        with open(Path.home()/'.pastaELN.temp', 'r', encoding='utf-8') as fTemp:
          content = json.loads(fTemp.read())
          if self.doc.get('_id', '') in content:
            del content[self.doc.get('_id', '')]
        with open(Path.home()/'.pastaELN.temp', 'w', encoding='utf-8') as fTemp:
          fTemp.write(json.dumps(content))
      if hasattr(self, 'key_-name'):
        self.doc['-name'] = getattr(self, 'key_-name').text().strip()
        if self.doc['-name'] == '':
          showMessage(self, 'Error', 'A created item has to have a valid name')
          return
        if self.doc['-type'][0]=='x0':  #prevent project-directory names that are identical
          others = self.comm.backend.db.getView('viewDocType/x0All')
          if '_id' in self.doc:
            others = [i for i in others if i['id']!=self.doc['_id']] # create list of names but filter own name
          others = [i['value'][0] for i in others]
          othersList = [createDirName(str(i),'x0', 0) for i in others] #create names
          while createDirName(self.doc['-name'],'x0', 0) in othersList:
            if re.search(r"_\d+$", self.doc['-name']) is None:
              self.doc['-name'] += '_1'
            else:
              self.doc['-name'] = '_'.join(self.doc['-name'].split('_')[:-1])+'_'+str(int(self.doc['-name'].split('_')[-1])+1)
      # loop through all the subitems
      for key in self.allKeys:
        valueOld = self.doc.get(key, '')
        if (key[0] in ['_', '-'] or key in ['image', 'metaVendor', 'metaUser']  #tags are already saved
            or not hasattr(self, f'key_{key}') and not hasattr(self, f'textEdit_{key}')):
          continue
        if key in ['comment','content']:
          text = getattr(self, f'textEdit_{key}').toPlainText().strip()
          if '_ids' not in self.doc or text:  #if group edit, text has to have text
            self.doc[key] = text
            if key == 'content' and '-branch' in self.doc:
              for branch in self.doc['-branch']:
                if branch['path'] is not None:
                  if branch['path'].endswith('.md'):
                    with open(self.comm.backend.basePath/branch['path'], 'w', encoding='utf-8') as fOut:
                      fOut.write(self.doc['content'])
                    logging.debug('Wrote new content to '+branch['path'])
                  else:
                    showMessage(self, 'Information', 'Did update the database but not the file on harddisk, since PASTA-ELN cannot write this format')
        elif isinstance(valueOld, list):  #items that are comma separated in the text-field
          self.doc[key] = getattr(self, f'key_{key}').text().strip().split(' ')
        elif isinstance(valueOld, str):
          if isinstance(getattr(self, f'key_{key}'), QComboBox):
            valueNew = getattr(self, f'key_{key}').currentText()
            dataNew  = getattr(self,f'key_{key}').currentData()
            if ((dataNew is not None and re.search(r"^[a-z\-]-[a-z0-9]{32}$",dataNew) is not None)
                or dataNew==''):
              #if docID is stored in currentData
              self.doc[key] = dataNew
            elif valueNew!='- no link -' or dataNew is None:
              self.doc[key] = valueNew
          else:                          #normal text field
            self.doc[key] = getattr(self, f'key_{key}').text().strip()
        elif valueOld is None and key in self.doc:  #important entry, set to empty string
          self.doc[key]=''
        else:
          print("**ERROR dialogForm unknown value type",key, valueOld)
      # new key-value pairs
      keyValueList = [self.keyValueListL.itemAt(i).widget().text() for i in range(self.keyValueListL.count())]# type: ignore[attr-defined]
      keyValueDict = dict(zip(keyValueList[::2],keyValueList[1::2] ))
      keyValueDict = {k:v for k,v in keyValueDict.items() if k}
      self.doc = keyValueDict | self.doc
      # ---- if project changed: only branch save; remaining data still needs saving
      newProjID = []
      if hasattr(self, 'projectComboBox') and self.projectComboBox.currentData() != '':
        parentPath = self.db.getDoc(self.projectComboBox.currentData())['-branch'][0]['path']
        if '_ids' in self.doc:  # group update
          for docID in self.doc['_ids']:
            doc = self.db.getDoc(docID)
            if doc['-branch'][0]['stack']!=self.projectComboBox.currentData(): #only if project changed
              if doc['-branch'][0]['path'] is None:
                newPath    = ''
              else:
                oldPath    = self.comm.backend.basePath/doc['-branch'][0]['path']
                newPath = f'{parentPath}/{oldPath.name}'
                oldPath.rename(self.comm.backend.basePath/newPath)
              self.db.updateBranch( doc['_id'], 0, 9999, [self.projectComboBox.currentData()], newPath)
        elif '-branch' in self.doc:             # sequential or single update
          if self.doc['-branch'][0]['stack']!=self.projectComboBox.currentData(): #only if project changed
            if self.doc['-branch'][0]['path'] is None:
              newPath    = ''
            else:
              oldPath = self.comm.backend.basePath/self.doc['-branch'][0]['path']
              newPath = f'{parentPath}/{oldPath.name}'
            self.db.updateBranch( self.doc['_id'], 0, 9999, [self.projectComboBox.currentData()], newPath)
            self.doc['-branch'][0] = {'stack':[self.projectComboBox.currentData()], 'path':newPath or None, 'child':9999, 'show':[True,True]}
        else:
          newProjID = [self.projectComboBox.currentData()]
      # ---- if docType changed: save; no further save to db required ----
      if hasattr(self, 'docTypeComboBox') and self.docTypeComboBox.currentData() != '':
        self.doc['-type'] = [self.docTypeComboBox.currentData()]
      if '_ids' in self.doc:                              # group update
        if '-name' in self.doc:
          del self.doc['-name']
        self.doc = {k:v for k,v in self.doc.items() if v} # filter out empty items
        for docID in self.doc.pop('_ids'):
          doc = self.db.getDoc(docID)
          doc.update( self.doc )
          self.comm.backend.editData(doc)
      elif '_id' in self.doc:                             # default update on item
        self.comm.backend.editData(self.doc)
      else:                                               # create new dataset
        self.comm.backend.addData(self.doc['-type'][0], copy.deepcopy(self.doc), newProjID)
      #!!! NO updates / redraw here since one does not know from where form came
      # e.g. sequential edit cannot have redraw here
      if command[0] is Command.FORM_SAVE_NEXT:
        for delKey in [i for i in self.doc.keys() if i[0] in ['-','_'] and i not in ['-name','-type','-tags']]:
          del self.doc[delKey]
        self.comm.changeTable.emit('', '')
      else:
        self.accept()  #close
        self.close()
    elif command[0] is Command.FORM_ADD_KV:
      self.keyValueLabel.show()
      self.keyValueListW.show()
      key = QLineEdit('')
      key.setPlaceholderText('key')
      key.setToolTip('Key (leave empty to delete key-value pair)')
      key.setValidator(QRegularExpressionValidator("[a-zA-Z0-9]\\S+"))
      value = QLineEdit('')
      value.setPlaceholderText('value')
      self.keyValueListL.addRow(key, value)
    elif command[0] is Command.FORM_SHOW_DOC:
      doc = copy.deepcopy(self.doc)
      if 'image' in doc:
        del doc['image']
      messageWindow = ScrollMessageBox('Details', doc, style='QScrollArea{min-width:600 px; min-height:400px}')
      ret = messageWindow.exec()
    else:
      print('**ERROR dialogForm: unknown Command ', command)
    return


  def textChanged(self) -> None:
    """
    Text changed in editor -> update the display on the right
    """
    key = self.sender().accessibleName()                                                                     # type: ignore
    getattr(self, f'textShow_{key}').setMarkdown(markdownStyler(
        getattr(self, f'textEdit_{key}').toPlainText()))
    return

  def delTag(self, _:str, tag:str) -> None:
    """
    Clicked button to delete tag
    """
    self.doc['-tags'].remove(tag)
    self.updateTagsBar()
    return

  def addTag(self, tag:Union[str,int]) -> None:
    """
    Clicked to add tag. Since one needs to use indexChanged to allow the user to enter text, that delivers a int. To allow to differentiate
    between both comboboxes, they cannot be the same (both int), hence grades has to be textChanged

    Args:
      tag (str, int): index (otherTags) or text (grades)
    """
    if isinstance(tag, str):#text from grades
      if tag!='':
        self.doc['-tags'] = [i for i in self.doc['-tags'] if i[0]!='_']
        self.doc['-tags'] += [f'_{len(tag)}']
        self.gradeChoices.setCurrentText('')
    elif tag<1:               #zero index from other-tags
      return
    else:
      tag = self.otherChoices.currentText()
      if tag not in self.doc['-tags']:
        self.doc['-tags'] += [tag]
      self.otherChoices.setCurrentText('')
    self.updateTagsBar()
    return

  def updateTagsBar(self) -> None:
    """
    After creation, tag removal, tag addition: update the information on screen
    """
    #update tags
    for i in reversed(range(self.tagsBarSubL.count())):
      self.tagsBarSubL.itemAt(i).widget().setParent(None)
    for tag in self.doc['-tags']:
      if tag in ['_curated']:
        continue
      if tag[0]=='_':
        Label('\u2605'*int(tag[1]), 'h3', self.tagsBarSubL, self.delTag, tag, 'click to remove')
      else:
        Label(tag, 'h3', self.tagsBarSubL, self.delTag, tag, 'click to remove')
    self.tagsBarSubL.addWidget(QWidget(), stretch=2)
    #update choices in combobox
    tagsAllList = self.comm.backend.db.getView('viewIdentify/viewTagsAll')
    tagsSet = {i['key'] for i in tagsAllList if i['key'][0]!='_'}
    newChoicesList = ['']+list(tagsSet.difference([i for i in self.doc['-tags'] if i[0]!='_']))
    self.otherChoices.clear()
    self.otherChoices.addItems(newChoicesList)
    return


  def changeTabs(self, idx:int) -> None:
    """
    Clicked on tabs to change to different one

    Args:
      idx (int): index of tab that was clicked
    """
    if idx == 0:
      self.btnAddKWPairs.show()
    else:
      self.btnAddKWPairs.hide()
    return


class Command(Enum):
  """ Commands used in this file """
  BUTTON_BAR       = 1
  CHANGE_EXTRACTOR = 2
  SAVE_IMAGE       = 3
  OPEN_FILEBROWSER = 4
  OPEN_EXTERNAL    = 5
  FOCUS_AREA       = 6
  FORM_SAVE        = 7
  FORM_CANCEL      = 8
  FORM_ADD_KV      = 9
  FORM_SHOW_DOC    = 10
  FORM_SAVE_NEXT   = 11



# Form implementation generated from reading ui file 'completed_uploads_base.ui'
#
# Created by: PyQt6 UI code generator 6.4.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PySide6 import QtCore, QtGui, QtWidgets


class Ui_CompletedUploadsForm(object):
  def setupUi(self, CompletedUploadsForm):
    CompletedUploadsForm.setObjectName("CompletedUploadsForm")
    CompletedUploadsForm.resize(1300, 475)
    self.gridLayout = QtWidgets.QGridLayout(CompletedUploadsForm)
    self.gridLayout.setObjectName("gridLayout")
    spacerItem = QtWidgets.QSpacerItem(20, 10, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Fixed)
    self.gridLayout.addItem(spacerItem, 1, 0, 1, 1)
    self.mainVerticalLayout = QtWidgets.QVBoxLayout()
    self.mainVerticalLayout.setObjectName("mainVerticalLayout")
    self.filterHorizontalLayout = QtWidgets.QHBoxLayout()
    self.filterHorizontalLayout.setObjectName("filterHorizontalLayout")
    self.filterTermLineEdit = QtWidgets.QLineEdit(parent=CompletedUploadsForm)
    self.filterTermLineEdit.setClearButtonEnabled(True)
    self.filterTermLineEdit.setObjectName("filterTermLineEdit")
    self.filterHorizontalLayout.addWidget(self.filterTermLineEdit)
    self.mainVerticalLayout.addLayout(self.filterHorizontalLayout)
    self.completedUploadsScrollArea = QtWidgets.QScrollArea(parent=CompletedUploadsForm)
    self.completedUploadsScrollArea.setWidgetResizable(True)
    self.completedUploadsScrollArea.setObjectName("completedUploadsScrollArea")
    self.scrollAreaWidgetContents = QtWidgets.QWidget()
    self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 1278, 404))
    self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
    self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
    self.verticalLayout_3.setObjectName("verticalLayout_3")
    self.completedUploadsVerticalLayout = QtWidgets.QVBoxLayout()
    self.completedUploadsVerticalLayout.setObjectName("completedUploadsVerticalLayout")
    self.verticalLayout_3.addLayout(self.completedUploadsVerticalLayout)
    self.completedUploadsScrollArea.setWidget(self.scrollAreaWidgetContents)
    self.mainVerticalLayout.addWidget(self.completedUploadsScrollArea)
    self.gridLayout.addLayout(self.mainVerticalLayout, 0, 0, 1, 1)

    self.retranslateUi(CompletedUploadsForm)
    QtCore.QMetaObject.connectSlotsByName(CompletedUploadsForm)

  def retranslateUi(self, CompletedUploadsForm):
    _translate = QtCore.QCoreApplication.translate
    CompletedUploadsForm.setWindowTitle(_translate("CompletedUploadsForm", "Dataverse upload history"))
    self.filterTermLineEdit.setToolTip(_translate("CompletedUploadsForm", "Enter project name / dataverse URL / finished time to filter the below listed tasks."))
    self.filterTermLineEdit.setPlaceholderText(_translate("CompletedUploadsForm", "Enter the information to filter the tasks."))
    self.completedUploadsScrollArea.setToolTip(_translate("CompletedUploadsForm", "<html><head/><body><p><span style=\" font-style:italic;\">Displays the history of finished dataverse uploads done in the past.</span></p></body></html>"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    CompletedUploadsForm = QtWidgets.QWidget()
    ui = Ui_CompletedUploadsForm()
    ui.setupUi(CompletedUploadsForm)
    CompletedUploadsForm.show()
    sys.exit(app.exec())



""" Contains utility functions. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: utils.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

import copy
import re
from asyncio import get_event_loop
from base64 import b64decode, b64encode
from json import dump, load
from logging import Logger
from os.path import exists, join
from pathlib import Path
from typing import Any, Callable, Type

from PySide6.QtCore import QSize
from PySide6.QtGui import QImage, QPixmap
from PySide6.QtWidgets import QBoxLayout
from cryptography.fernet import Fernet, InvalidToken
from qtawesome import icon

from pasta_eln.dataverse.client import DataverseClient
from pasta_eln.dataverse.config_error import ConfigError
from pasta_eln.dataverse.upload_status_values import UploadStatusValues
from pasta_eln.miscTools import upOut


def update_status(status: str,
                  status_label_set_text_callback: Callable[[str], None],
                  status_icon_set_pixmap_callback: Callable[[QPixmap | QImage | str], None]) -> None:
  """
  Updates the status with the corresponding label and icon.

  Args:
      status (str): The status value.
      status_label_set_text_callback (Callable[[str], None]): The callback function to set the status label text.
      status_icon_set_pixmap_callback (Callable[[QPixmap | QImage | str], None]): The callback function to set the status icon pixmap.

  Explanation:
      This function updates the status with the corresponding label and icon based on the given status value.
      It uses the provided callback functions to set the status label text and status icon pixmap.
  """
  match status:
    case UploadStatusValues.Queued.name:
      icon_name = 'fa.circle-o-notch'
    case UploadStatusValues.Uploading.name:
      icon_name = 'fa.cloud-upload'
    case UploadStatusValues.Cancelled.name:
      icon_name = 'fa.minus-circle'
    case UploadStatusValues.Finished.name:
      icon_name = 'fa.check-circle-o'
    case UploadStatusValues.Error.name:
      icon_name = 'fa.times-circle-o'
    case UploadStatusValues.Warning.name:
      icon_name = 'fa.warning'
    case _:
      icon_name = 'fa.times-circle-o'
  status_label_set_text_callback(status)
  status_icon_set_pixmap_callback(icon(icon_name).pixmap(QSize(30, 30)))


def set_authors(logger: Logger, metadata: dict[str, Any]) -> None:
  """
  Sets the authors in the metadata.

  Explanation:
      This function sets the authors in the metadata based on the configuration file.
      It retrieves the authors from the configuration file and updates the metadata accordingly.

  Args:
      logger (Logger): The logger instance for logging errors.
      metadata (dict[str, Any]): The metadata dictionary to update.
  """
  config = read_pasta_config_file(logger)
  if config is None:
    raise log_and_create_error(logger, ConfigError, "Config file not found, Corrupt installation!")
  if 'authors' not in config:
    raise log_and_create_error(logger, ConfigError, "Incorrect config file, authors not found!")
  author_field = next(
    f for f in metadata['datasetVersion']['metadataBlocks']['citation']['fields'] if f['typeName'] == 'author')
  authors_list = author_field.get('value')
  if authors_list is None:
    raise log_and_create_error(logger, ConfigError, "Incorrect config file, authors value should be found!")
  template_list = author_field.get('valueTemplate', [{}])
  template = template_list[0] if isinstance(template_list, list) and template_list else {}
  authors_list.clear()
  for author in config['authors']:
    author_copy = copy.deepcopy(template)
    author_copy['authorName']['value'] = ', '.join(filter(None, [author['last'], author['first']]))
    author_copy['authorIdentifierScheme']['value'] = "ORCID"
    author_copy['authorIdentifier']['value'] = author['orcid']
    author_copy['authorAffiliation']['value'] = ', '.join([o['organization'] for o in author['organizations']])
    authors_list.append(author_copy)


def get_flattened_metadata(metadata: dict[str, Any]) -> dict[str, Any]:
  """
  Returns a flattened version of the metadata.

  Args:
      metadata (dict[str, Any]): The metadata to be flattened.

  Returns:
      dict[str, Any]: The flattened metadata.

  Explanation:
      This function takes a nested metadata dictionary and returns a flattened version of the metadata.
      It extracts the values from the nested structure and creates a new dictionary with a flattened structure.
      The resulting dictionary contains the field names as keys and their corresponding values.

  Exceptions:
      KeyError: If a field name is not found in the metadata dictionary.
  """
  adjusted_metadata = {}
  if metadata['datasetVersion']['license']:
    adjusted_metadata['license'] = metadata['datasetVersion']['license']
  for _, metablock in metadata['datasetVersion']['metadataBlocks'].items():
    for field in metablock['fields']:
      if field['value']:
        adjusted_metadata[field['typeName']] = field['value']
  return adjusted_metadata


def set_template_values(logger: Logger, metadata: dict[str, Any]) -> None:
  """
  Set template values in the metadata.

  This function sets template values in the metadata dictionary based on the type of each field.
  If the metadata is empty, a warning is logged.
  The function iterates through the metadata blocks
  and fields, and for each field, it matches the type class and performs the following actions:
  - For "primitive" or "controlledVocabulary" types:
      - If the field is a multiple type, the value template is set to a copy of the value, and the value is cleared.
      - If the field is a single type, the value template is set to the value, and the value is set to an empty string.
  - For "compound" types:
      - The value template is set to a copy of the value, and the value is cleared.
  - For unknown type classes, a warning is logged.

  Args:
      logger (Logger): The logger object used for logging warnings.
      metadata (dict[str, Any]): The metadata dictionary to update.

  Examples:
      >>> logger = Logger()
      >>> metadata = {'datasetVersion': {'metadataBlocks': {'citation': {'fields': [{'typeClass': 'primitive', 'multiple': False, 'value': 'Example'}]}}}}
      >>> set_template_values(logger, metadata)
      >>> metadata
      {'datasetVersion': {'metadataBlocks': {'citation': {'fields': [{'typeClass': 'primitive', 'multiple': False, 'value': '', 'valueTemplate': 'Example'}]}}}}
  """
  if not metadata:
    logger.warning("Empty metadata, make sure the metadata is loaded correctly...")
    return
  for _, metablock in metadata['datasetVersion']['metadataBlocks'].items():
    for field in metablock['fields']:
      match field['typeClass']:
        case "primitive" | "controlledVocabulary":
          # Array of primitive types
          if field['multiple']:
            field['valueTemplate'] = field['value'].copy()
            field['value'].clear()
          # Single primitive type, set to empty string
          else:
            field['valueTemplate'] = field['value']
            field['value'] = ""
        case "compound":
          field['valueTemplate'] = field['value'].copy()
          field['value'].clear()
        case _:
          logger.warning(f"Unsupported type class: {field['typeClass']}")


def check_if_minimal_metadata_exists(logger: Logger,
                                     metadata: dict[str, Any],
                                     check_title: bool = True) -> dict[str, list[str]] | None:
  """
  Checks if the minimal metadata exists and returns the missing information.

  Args:
      logger (Logger): The logger object for logging warnings.
      metadata (dict[str, Any]): The metadata dictionary.
      check_title (bool, optional): Flag to indicate if the title should be checked.
      Defaults to True.

  Returns:
      dict[str, list[str]] | None: The missing information dictionary or None if metadata is empty.
      missing_information = {
                              'title': [],
                              'author': [],
                              'datasetContact': [],
                              'dsDescription': [],
                              'subject': []
                            }

  """
  if not metadata:
    logger.warning("Empty metadata, make sure the metadata is loaded correctly...")
    return None
  missing_information: dict[str, list[str]] = {
    'title': [],
    'author': [],
    'datasetContact': [],
    'dsDescription': [],
    'subject': []
  }
  check_if_field_value_not_null(get_citation_field(metadata, 'title'),
                                missing_information,
                                "title",
                                check_title)
  check_if_field_value_not_null(get_citation_field(metadata, 'subject'),
                                missing_information,
                                "subject")
  check_if_compound_field_value_is_missing(
    get_citation_field(metadata, 'author'),
    'author',
    missing_information,
    [('authorName', 'Author Name'),
     ('authorAffiliation', 'Author Affiliation'),
     ('authorIdentifierScheme', 'Author Identifier Scheme'),
     ('authorIdentifier', 'Author Identifier')])
  check_if_compound_field_value_is_missing(
    get_citation_field(metadata, 'datasetContact'),
    'datasetContact',
    missing_information,
    [('datasetContactName', 'Dataset Contact Name'),
     ('datasetContactAffiliation', 'Dataset Contact Affiliation'),
     ('datasetContactEmail', 'Dataset Contact Email')])
  check_if_compound_field_value_is_missing(
    get_citation_field(metadata, 'dsDescription'),
    'dsDescription',
    missing_information,
    [('dsDescriptionValue', 'Dataset Description Value'),
     ('dsDescriptionDate', 'Dataset Description Date')])
  return missing_information


def get_citation_field(metadata: dict[str, Any], name: str) -> dict[str, Any]:
  """
  Retrieves the citation field from the metadata.

  Args:
      metadata (dict[str, Any]): The metadata dictionary.
      name (str): The name of the field to retrieve.

  Returns:
      dict[str, Any]: The citation field.

  """
  return next(
    f for f in metadata['datasetVersion']['metadataBlocks']['citation']['fields'] if f.get('typeName') == name)


def check_if_compound_field_value_is_missing(field: dict[str, Any],
                                             field_key: str,
                                             missing_information: dict[str, list[str]],
                                             sub_fields: list[tuple[str, str]]) -> None:
  """
  Checks if the compound field value is missing and updates the missing_information dictionary if it is.

  Args:
      field (dict[str, Any]): The compound field dictionary.
      field_key (str): The key representing the compound field.
      missing_information (dict[str, list[str]]): The dictionary to track missing information.
      sub_fields (list[tuple[str, str]]): The list of sub-fields to check.
  """
  check_if_field_value_not_null(field, missing_information, field_key)
  if values := field.get('value'):
    for field_value in values:
      for field_name, missing_field_name in sub_fields:
        check_if_field_value_is_missing(field_value, field_key, field_name, missing_field_name,
                                        missing_information)


def check_if_field_value_is_missing(field: dict[str, Any],
                                    field_key: str,
                                    field_name: str,
                                    missing_field_name: str,
                                    missing_information: dict[str, list[str]]) -> None:
  """
  Checks if the field value is missing and updates the missing_information dictionary if it is.

  Args:
      field (dict[str, Any]): The field dictionary.
      field_key (str): The key representing the field.
      field_name (str): The name of the field.
      missing_field_name (str): The name of the missing field.
      missing_information (dict[str, list[str]]): The dictionary to track missing information.

  """
  if not field.get(field_name, {}).get('value'):
    missing_message = f"{missing_field_name} field is missing for one of the {field_key}s!"
    if missing_message not in missing_information[field_key]:
      missing_information[field_key].append(missing_message)


def check_if_field_value_not_null(field: dict[str, Any],
                                  missing_information: dict[str, list[str]],
                                  missing_field_name: str,
                                  check: bool = True) -> None:
  """
  Checks if the field value is not null and updates the missing_information dictionary if it is.

  Args:
      field (dict[str, Any]): The field dictionary.
      missing_information (dict[str, list[str]]): The dictionary to track missing information.
      missing_field_name (str): The name of the missing field.
      check (bool, optional): Flag to indicate if the check should be performed. Defaults to True.

  """
  if check and not field.get('value'):
    missing_message = f"Add at-least a single entry for {missing_field_name.capitalize()}!"
    if missing_message not in missing_information[missing_field_name]:
      missing_information[missing_field_name].append(missing_message)


def get_encrypt_key(logger: Logger) -> tuple[bool, bytes]:
  """
  Gets the dataverse encryption key.

  Explanation:
  This function retrieves the dataverse encryption key from the configuration file. If the key does not exist, a new key is generated and stored in the configuration file.

  Args:
      logger (Logger): The logger instance for logging messages.

  Raises:
    EnvironmentError: If the config file is not found or corrupt.

  Returns:
      tuple[bool, bytes]: A tuple containing a boolean indicating whether the key exists and the encryption key itself.
  """
  logger.info("Getting dataverse encrypt key..")
  config = read_pasta_config_file(logger)
  key_exists = False
  if 'dataverseEncryptKey' not in config:
    logger.warning("Dataverse encrypt key does not exist, hence generating a new key..")
    config['dataverseEncryptKey'] = b64encode(Fernet.generate_key()).decode('ascii')
    write_pasta_config_file(logger, config)
  else:
    key_exists = True
  key = b64decode(config['dataverseEncryptKey'])
  return key_exists, key


def encrypt_data(logger: Logger, encrypt_key: bytes | None, data: str | None) -> str | None:
  """
  Encrypts the provided data using the given encryption key.

  Explanation:
      This function encrypts the provided data using the given encryption key.
      It returns the encrypted data as a string.

  Args:
      logger (Logger): The logger instance for logging information.
      encrypt_key (bytes | None): The encryption key used for encryption.
      data (str | None): The data to be encrypted.

  Returns:
      str | None: The encrypted data as a string, or None if either the encrypt_key or data is None.
  """
  if encrypt_key is None or data is None:
    logger.warning("encrypt_key/data cannot be None")
    return None
  try:
    fernet = Fernet(encrypt_key)
    data = fernet.encrypt(data.encode('ascii')).decode('ascii')
  except InvalidToken as e:
    logger.error("Invalid token: %s", e)
    return None
  except ValueError as e:
    logger.error("Value error: %s", e)
    return None
  except AttributeError as e:
    logger.error("AttributeError: %s", e)
    return None
  return data


def decrypt_data(logger: Logger, encrypt_key: bytes | None, data: str | None) -> str | None:
  """
  Decrypts the provided data using the given encryption key.

  Explanation:
      This function decrypts the provided data using the given encryption key.
      It returns the decrypted data as a string.

  Args:
      logger (Logger): The logger instance for logging information.
      encrypt_key (bytes | None): The encryption key used for decryption.
      data (str | None): The data to be decrypted.

  Returns:
      str | None: The decrypted data as a string, or None if either the encrypt_key or data is None.
  """
  if encrypt_key is None or data is None:
    logger.warning("encrypt_key/data cannot be None")
    return None
  try:
    fernet = Fernet(encrypt_key)
    data = fernet.decrypt(data.encode('ascii')).decode('ascii')
  except InvalidToken as e:
    logger.error("Invalid token: %s", e)
    return None
  except ValueError as e:
    logger.error("Value error: %s", e)
    return None
  except AttributeError as e:
    logger.error("AttributeError: %s", e)
    return None
  return data


def read_pasta_config_file(logger: Logger) -> dict[str, Any]:
  """
  Reads the PASTA configuration file.

  Explanation:
      This function reads the PASTA configuration file and returns the configuration as a dictionary.
      If the configuration file is not found, it raises a DatabaseError.

  Args:
      logger (Logger): The logger instance for logging errors.

  Returns:
      dict[str, Any] | None: The configuration dictionary, or None if the configuration file is not found.
  """
  config_file_name = join(Path.home(), '.pastaELN.json')
  if not exists(config_file_name):
    raise log_and_create_error(logger, ConfigError, "Config file not found, Corrupt installation!")
  logger.info("Reading config file: %s", config_file_name)
  with open(config_file_name, 'r', encoding='utf-8') as confFile:
    config = load(confFile)
  return config


def write_pasta_config_file(logger: Logger, config_data: dict[str, Any]) -> None:
  """
  Writes the PASTA config file with the provided configuration data.

  Explanation:
      This function writes the PASTA config file with the provided configuration data.
      It raises a DatabaseError if the config file does not exist.

  Args:
      logger (Logger): The logger instance for logging information.
      config_data (dict[str, Any]): The configuration data to be written to the config file.
  """
  config_file_name = join(Path.home(), '.pastaELN.json')
  if not exists(config_file_name):
    raise log_and_create_error(logger, ConfigError, "Config file not found, Corrupt installation!")
  logger.info("Writing config file: %s", config_file_name)
  with open(config_file_name, 'w', encoding='utf-8') as confFile:
    dump(config_data, confFile, ensure_ascii=False, indent=4)


def log_and_create_error(logger: Logger, exception_type: Type[Exception], error_message: str) -> Exception:
  """
  Logs an error message and creates an exception.

  Explanation:
      This function logs the provided error message and creates an exception of the specified type.

  Args:
      logger (Logger): The logger instance for logging errors.
      exception_type (Type[Exception]): The type of exception to create.
      error_message (str): The error message to log and include in the exception.

  Returns:
      Exception: The created exception.
  """
  logger.error(error_message)
  return exception_type(error_message)


def check_login_credentials(logger: Logger, api_token: str, server_url: str) -> tuple[bool, str]:
  """
  Checks if the login credentials (API token and server URL) are valid.

  Explanation:
      This function checks if the provided API token and server URL are valid by
      attempting to reach the dataverse server and verifying the token.

  Args:
      logger (Logger): The logger instance for logging information.
      api_token (str): The API token for authentication.
      server_url (str): The URL of the dataverse server.

  Returns:
      tuple[bool, str]: A tuple containing a boolean indicating if the login credentials are valid
      and a message describing the result.
  """
  logger.info("Checking if login info is valid, server_url: %s", server_url)
  dataverse_client = DataverseClient(server_url, api_token)
  event_loop = get_event_loop()
  success, message = event_loop.run_until_complete(dataverse_client.check_if_dataverse_server_reachable())
  if success:
    token_valid = event_loop.run_until_complete(dataverse_client.check_if_api_token_is_valid())
    if token_valid:
      result = True, "Data server is reachable and token is valid"
    else:
      result = False, "Data server is reachable but token is invalid"
      logger.warning("Data server is reachable but API token is invalid!")
  elif 'Unauthorized' in message:
    result = False, "Data server is reachable but API token is invalid"
    logger.warning("Data server is reachable but API token is invalid: %s", message)
  else:
    result = False, "Data server is not reachable"
    logger.warning("Data server is not reachable: %s", message)
  return result


def check_if_dataverse_exists(logger: Logger, api_token: str, server_url: str, dataverse_id: str) -> bool:
  """
  Checks if a Dataverse exists based on the provided server URL, API token, and Dataverse ID.

  Args:
      logger (Logger): The logger object for logging information.
      api_token (str): The API token for authentication.
      server_url (str): The URL of the Dataverse server.
      dataverse_id (str): The ID of the Dataverse to check existence for.

  Returns:
      bool: True if the Dataverse exists, False otherwise.
  """
  logger.info("Checking if login info is valid, server_url: %s", server_url)
  dataverse_client = DataverseClient(server_url, api_token)
  event_loop = get_event_loop()
  info = event_loop.run_until_complete(dataverse_client.get_dataverse_info(dataverse_id))
  result = False
  if isinstance(info, dict):
    result = (info.get("id", None) is not None
              and info.get("alias", None) == dataverse_id)
  else:
    logger.warning("Data verse with id %s does not exist, Server message: %s", dataverse_id, info)
  return result


def adjust_type_name(camel_case_string: str) -> str:
  """
  Adjusts the type name from camel case to title case.

  Splits the camel case string into individual words and capitalizes the first letter of each word.
  If a word starts with an uppercase letter, it is left as is.

  Args:
      camel_case_string (str): The input string in camel case.

  Returns:
      str: The adjusted type name in title case.
  """
  split = re.findall(r'[A-Z]?[a-z]+|[A-Z]+(?=[A-Z]|$)',
                     camel_case_string)
  return ' '.join([i.capitalize() if i[0].islower() else i for i in split])


def clear_value(items: dict[str, Any] | None = None) -> None:
  """
  Clears the 'value' attribute of each item in the given dictionary.

  Args:
      items(dict[str, Any] | None): The dictionary containing items to be cleared.

  Explanation:
      This function iterates over the values of the given dictionary and sets the 'value'
      attribute to None for each item.
      If the 'items' argument is None, the function returns immediately.

  """
  if items is None:
    return
  for item in items.values():
    if item and "value" in item:
      item["value"] = None


def is_date_time_type(type_name: str) -> bool:
  """
  Checks if the given type_name contains 'date' or 'time' as a substring.

  Args:
      type_name (str): The type name to check.

  Returns:
      bool: True if the type_name contains 'date' or 'time', False otherwise.

  Explanation:
      This function checks if the given type_name contains 'date' or 'time' as a substring.
      It performs a case-insensitive search to determine the presence of 'date' or 'time' in the type_name.

  """
  return any(
    map(type_name.lower().__contains__, ["date", "time"]))


def delete_layout_and_contents(layout: QBoxLayout) -> None:
  """
  Deletes the layout and its contents.

  Args:
      layout (QBoxLayout): The layout to delete.

  Explanation:
      This function deletes the given layout and its contents.
      It iterates over the widgets in the layout and removes them from their parent.
      Finally, it sets the layout's parent to None.

  """
  if layout is None:
    return
  for widget_pos in reversed(range(layout.count())):
    if item := layout.itemAt(widget_pos):
      item.widget().setParent(None)
  layout.setParent(None)


def get_formatted_message(missing_metadata: dict[str, list[str]]) -> str:
  """
  Returns a formatted message with missing metadata information.

  Args:
      missing_metadata (dict[str, list[str]]): A dictionary containing missing metadata information.

  Returns:
      str: A formatted message with the missing metadata information.

  Explanation:
      This function takes a dictionary of missing metadata information and returns a formatted message
      with the missing information. The message is formatted as an HTML string and includes a list of missing
      metadata items grouped by their corresponding metadata names.
  """
  if not any(missing_metadata.values()):
    return ""
  name_mapping = {
    'author': 'Author',
    'datasetContact': "Dataset Contact",
    'dsDescription': "Dataset Description",
    'subject': "Subject"
  }
  message = (
    "<html><p><i>Goto 'Edit Metadata' dialog, select 'Minimal' metadata list, enter the below given missing information and retry the upload!</i></p>"
  )
  for metadata_name, missing_list in missing_metadata.items():
    if missing_list:
      message += f"<br></br><b><i>{name_mapping[metadata_name]}:</i></b><ul>"
      for missing in missing_list:
        message += f"<i style=\"color:Crimson\"><li>{missing}</li></i>"
      message += "</ul>"
  message += "</html>"
  return message


def get_formatted_metadata_message(metadata: dict[str, Any]) -> str:
  """
  Formats metadata into an HTML message for display.

  Args:
      metadata (dict): A dictionary containing metadata information.

  Returns:
      str: An HTML-formatted message representing the metadata.

  Examples:
      >>> metadata = {
      ...     'datasetVersion': {
      ...         'license': {'name': 'CC0', 'uri': 'https://creativecommons.org/publicdomain/zero/1.0/'},
      ...         'metadataBlocks': {
      ...             'citation': {
      ...                 'displayName': 'Citation Metadata',
      ...                 'fields': [
      ...                     {'typeName': 'title', 'value': 'Test Title', 'multiple': False, 'typeClass': 'primitive'}
      ...                 ]
      ...             }
      ...         }
      ...     }
      ... }
      >>> get_formatted_metadata_message(metadata)
      '<html>...</html>'
  """

  message = "<html>"
  if metadata['datasetVersion']['license']:
    message += "<b style=\"color:Black\">License Metadata:</b><ul>"
    message += f"<li style=\"color:Gray\">Name: <i>{metadata['datasetVersion']['license']['name']}</i></li>"
    message += f"<li style=\"color:Gray\">URI: <i>{metadata['datasetVersion']['license']['uri']}</i></li>"
    message += "</ul>"
  for _, metablock in metadata['datasetVersion']['metadataBlocks'].items():
    message += f"<b style=\"color:Black\">{metablock['displayName']}:</b><ul>"
    has_value = False
    for field in metablock['fields']:
      if field['value']:
        has_value = has_value or len(field['value']) > 0
        message += f"<b style=\"color:#737373\"><li>{adjust_type_name(field['typeName'])}:</li></b><ul>"
        if field["multiple"] and field["typeClass"] == "compound":
          for index, field_value in enumerate(field['value']):
            message += f"<li style=\"color:Gray\">Item {index + 1}:</li><ul>"
            for _, value in field_value.items():
              message += f"<li style=\"color:Gray\">{adjust_type_name(value['typeName'])}: <i>{value['value']}</li></i>"
            message += "</ul>"
          message += "</ul>"
        elif field["multiple"] and field["typeClass"] == "controlledVocabulary":
          for field_value in field['value']:
            message += f"<i style=\"color:Gray\"><li>{field_value}</li></i>"
          message += "</ul>"
        elif field["multiple"] and field["typeClass"] == "primitive":
          for field_value in field['value']:
            message += f"<i style=\"color:Gray\"><li>{field_value}</li></i>"
          message += "</ul>"
        else:
          message += f"<i style=\"color:Gray\"><li>{field['value']}</li></i></ul>"
    if not has_value:
      message += "<li style=\"color:#737373\">No value</li></ul>"
    message += "</ul>"
  message += "</html>"
  return message


def get_formatted_dataverse_url(dataverse_url: str) -> str:
  """
  Formats a Dataverse URL.

  Explanation:
      This function takes a Dataverse URL as input and extracts the persistent ID from it.
      It then formats the URL with the persistent ID in an HTML structure for display.

  Args:
      dataverse_url (str): The Dataverse URL to be formatted.

  Returns:
      str: The formatted Dataverse URL in HTML structure.
  """

  formatted_dataverse_url = ""
  if dataverse_url:
    if "persistentId=" not in dataverse_url:
      return formatted_dataverse_url
    persistent_id = dataverse_url.split("persistentId=", 1)[1]
    formatted_dataverse_url = (
      f"<html><head/><body><p>Dataverse URL: "
      f"<a href='{dataverse_url}'>"
      f"<span style='font-style:italic; text-decoration: underline; color:#77767b;'>{persistent_id}</span>"
      f"</a></p></body></html>")
  return formatted_dataverse_url


def get_data_hierarchy_types(data_hierarchy: dict[str, Any] | None) -> list[str | Any]:
  """
  Gets the types of data hierarchy.

  Explanation:
      This function extracts the types of data hierarchy from the provided data hierarchy dictionary.
      It filters out specific types and appends 'Unidentified' to the list of data hierarchy types.

  Args:
      data_hierarchy (dict[str, Any] | None): The data hierarchy dictionary.

  Returns:
      list[str | Any]: The list of data hierarchy types.
  """
  if data_hierarchy is None:
    return []
  data_hierarchy_types = []
  for data_type in data_hierarchy:
    if (data_type and not data_type.isspace()
        and data_type not in ("x0", "x1", "x2")):
      type_capitalized = data_type.capitalize()
      if type_capitalized not in data_hierarchy_types:
        data_hierarchy_types.append(type_capitalized)
  data_hierarchy_types.append("Unidentified")
  return data_hierarchy_types


def get_db_credentials(logger: Logger) -> dict[str, str]:
  """
  Gets the database credentials.

  Explanation:
      This function retrieves the database credentials from the configuration file based on the default project group.
      It validates the configuration and returns the database credentials if found.

  Args:
      logger (Logger): The logger instance for logging messages.

  Returns:
      dict[str, str]: A dictionary containing the database credentials for the default project group.
      {
        'db_name': db_name,
        'username': username,
        'password': password,
      }

  Raises:
      ConfigError: If the configuration file is missing or incorrect entries in the config file.
  """

  config = read_pasta_config_file(logger)
  def_project_group_name = config['defaultProjectGroup']
  project_groups = config.get('projectGroups')
  if not def_project_group_name or not project_groups:
    raise log_and_create_error(logger, ConfigError,
                               "Incorrect config file, defaultProjectGroup/projectGroups not found!")
  main_group = project_groups.get(def_project_group_name)
  if not main_group:
    raise log_and_create_error(logger, ConfigError,
                               "Incorrect config file, defaultProjectGroup not found in projectGroups!")
  local_info = main_group.get('local')
  if not local_info:
    raise log_and_create_error(logger, ConfigError,
                               "Incorrect config file, user or password not found!")
  db_name = local_info.get('database')
  cred = local_info.get('cred')
  if not db_name:
    raise log_and_create_error(logger, ConfigError,
                               "Incorrect config file, database name not found!")
  if cred:
    username, password = decrypt_credentials(cred)
  else:
    username = local_info.get('user')
    password = local_info.get('password')
  if username and password:
    return {
      'username': username,
      'password': password,
      'db_name': db_name
    }
  else:
    raise log_and_create_error(logger, ConfigError,
                               "Incorrect config file, user/password/cred not found for the default project group!")


def decrypt_credentials(cred: str) -> tuple[str, str]:
  """
  Decrypts credentials.

  Explanation:
      This function decrypts the provided credentials and extracts the username and password.
      It splits the decrypted credentials by ':' and returns the username and password as a tuple.

  Args:
      cred (str): The encrypted credentials to decrypt.

  Returns:
      tuple[str, str]: A tuple containing the decrypted username and password.
  """
  user_name = ''
  pass_word = ''
  if cred:
    decrypted_creds = upOut(cred)[0].split(':')
    user_name = decrypted_creds[0]
    pass_word = decrypted_creds[1]
  return user_name, pass_word



""" Graphical user interface includes all widgets """
import json
import logging
import os
import sys
import webbrowser
from enum import Enum
from pathlib import Path
from typing import Any, Union

from PySide6.QtCore import QCoreApplication, Slot  # pylint: disable=no-name-in-module
from PySide6.QtGui import QIcon, QPixmap, QShortcut  # pylint: disable=no-name-in-module
from PySide6.QtWidgets import QApplication, QFileDialog, QMainWindow  # pylint: disable=no-name-in-module
from qt_material import apply_stylesheet  # of https://github.com/UN-GCPDS/qt-material

from pasta_eln import __version__
from pasta_eln.GUI.dataverse.config_dialog import ConfigDialog
from pasta_eln.GUI.dataverse.main_dialog import MainDialog
from .GUI.body import Body
from .GUI.config import Configuration
from .GUI.data_hierarchy.data_hierarchy_editor_dialog import DataHierarchyEditorDialog
from .GUI.form import Form
from .GUI.projectGroup import ProjectGroup
from .GUI.sidebar import Sidebar
from .backend import Backend
from .fixedStringsJson import shortcuts
from .guiCommunicate import Communicate
from .guiStyle import Action, ScrollMessageBox, showMessage, widgetAndLayout
from .inputOutput import exportELN
from .miscTools import restart, updateExtractorList

os.environ['QT_API'] = 'pyside6'


# Subclass QMainWindow to customize your application's main window
class MainWindow(QMainWindow):
  """ Graphical user interface includes all widgets """

  def __init__(self) -> None:
    # global setting
    super().__init__()
    venv = ' without venv' if sys.prefix == sys.base_prefix and 'CONDA_PREFIX' not in os.environ else ' in venv'
    self.setWindowTitle(f"PASTA-ELN {__version__}{venv}")
    self.resize(self.screen().size()) #self.setWindowState(Qt.Window Maximized) https://bugreports.qt.io/browse/PYSIDE-2706 https://bugreports.qt.io/browse/QTBUG-124892
    resourcesDir = Path(__file__).parent / 'Resources'
    self.setWindowIcon(QIcon(QPixmap(resourcesDir / 'Icons' / 'favicon64.png')))
    self.backend = Backend()
    self.comm = Communicate(self.backend)
    self.comm.formDoc.connect(self.formDoc)
    self.dataverseMainDialog: MainDialog | None = None
    self.dataverseConfig: ConfigDialog | None = None

    # Menubar
    menu = self.menuBar()
    projectMenu = menu.addMenu("&Project")
    Action('&Export project to .eln',        self, [Command.EXPORT],         projectMenu)
    Action('&Import .eln',                   self, [Command.IMPORT],         projectMenu, shortcut='Ctrl+A')
    projectMenu.addSeparator()
    Action('&Export all projects to .eln',   self, [Command.EXPORT_ALL],     projectMenu)
    Action('&Exit',                          self, [Command.EXIT],           projectMenu)

    viewMenu = menu.addMenu("&Lists")
    if hasattr(self.backend, 'db'):
      for docType, docLabel in self.comm.backend.db.dataLabels.items():
        if docType[0] == 'x' and docType[1] != '0':
          continue
        shortcut = self.comm.backend.db.dataHierarchy[docType].get('shortcut','')
        shortcut = None if shortcut=='' else f"Ctrl+{shortcut}"
        Action(docLabel,                     self, [Command.VIEW, docType],  viewMenu, shortcut=shortcut)
        if docType == 'x0':
          viewMenu.addSeparator()
      viewMenu.addSeparator()
      Action('&Tags',                        self, [Command.VIEW, '_tags_'], viewMenu, shortcut='Ctrl+T')
      Action('&Unidentified',                self, [Command.VIEW, '-'],      viewMenu, shortcut='Ctrl+U')

    systemMenu = menu.addMenu("&System")
    Action('&Project groups',                self, [Command.PROJECT_GROUP],   systemMenu)
    changeProjectGroups = systemMenu.addMenu("&Change project group")
    if hasattr(self.backend, 'configuration'):                            # not case in fresh install
      for name in self.backend.configuration['projectGroups'].keys():
        Action(name,                         self, [Command.CHANGE_PG, name], changeProjectGroups)
    Action('&Synchronize',                   self, [Command.SYNC],            systemMenu, shortcut='F5')
    Action('&Data hierarchy editor',         self, [Command.DATAHIERARCHY],        systemMenu, shortcut='F8')
    Action('&Dataverse Configuration',         self, [Command.DATAVERSE_CONFIG],        systemMenu, shortcut='F10')
    Action('&Upload to Dataverse',         self, [Command.DATAVERSE_MAIN],        systemMenu, shortcut='F11')
    systemMenu.addSeparator()
    Action('&Test extraction from a file',   self, [Command.TEST1],           systemMenu)
    Action('Test &selected item extraction', self, [Command.TEST2],           systemMenu, shortcut='F2')
    Action('Update &Extractor list',         self, [Command.UPDATE],          systemMenu)
    systemMenu.addSeparator()
    Action('&Configuration',                 self, [Command.CONFIG],          systemMenu, shortcut='Ctrl+0')

    helpMenu = menu.addMenu("&Help")
    Action('&Website',                       self, [Command.WEBSITE],         helpMenu)
    Action('&Verify database',               self, [Command.VERIFY_DB],       helpMenu, shortcut='Ctrl+?')
    Action('Shortcuts',                      self, [Command.SHORTCUTS],       helpMenu)
    # shortcuts for advanced usage (user should not need)
    QShortcut('F9', self, lambda: self.execute([Command.RESTART]))

    # GUI elements
    mainWidget, mainLayout = widgetAndLayout('H')
    self.setCentralWidget(mainWidget)  # Set the central widget of the Window
    body = Body(self.comm)  # body with information
    self.sidebar = Sidebar(self.comm)  # sidebar with buttons
    # sidebarScroll = QScrollArea()
    # sidebarScroll.setWidget(self.sidebar)
    # if hasattr(self.comm.backend, 'configuration'):
    #   sidebarScroll.setFixedWidth(self.comm.backend.configuration['GUI']['sidebarWidth']+10)
    # mainLayout.addWidget(sidebarScroll)
    mainLayout.addWidget(self.sidebar)
    mainLayout.addWidget(body)
    self.comm.changeTable.emit('x0', '')


  @Slot(str)
  def formDoc(self, doc: dict[str, Any]) -> None:
    """
    What happens when new/edit dialog is shown

    Args:
      doc (dict): document
    """
    if '_id' in doc:
      logging.debug('gui:formdoc ' + str(doc['_id']))
    elif '_ids' in doc:
      logging.debug('gui:formdoc ' + str(doc['_ids']))
    else:
      logging.debug('gui:formdoc of type ' + str(doc['-type']))
    formWindow = Form(self.comm, doc)
    ret = formWindow.exec()
    if ret == 0:
      self.comm.stopSequentialEdit.emit()
    return


  def execute(self, command: list[Any]) -> None:
    """
    action after clicking menu item
    """
    # file menu
    if command[0] is Command.EXPORT:
      if self.comm.projectID == '':
        showMessage(self, 'Error', 'You have to open a project to export', 'Warning')
        return
      fileName = QFileDialog.getSaveFileName(self, 'Save project into .eln file', str(Path.home()), '*.eln')[0]
      if fileName != '' and hasattr(self.backend, 'db'):
        docTypes = [i for i in self.comm.backend.db.dataLabels.keys() if i[0]!='x']
        status = exportELN(self.comm.backend, [self.comm.projectID], fileName, docTypes)
        showMessage(self, 'Finished', status, 'Information')
    elif command[0] is Command.EXPORT_ALL:
      fileName = QFileDialog.getSaveFileName(self, 'Save everything to .eln file', str(Path.home()), '*.eln')[0]
      if fileName != '' and hasattr(self.backend, 'db'):
        docTypes = [i for i in self.comm.backend.db.dataLabels.keys() if i[0]!='x']
        allProjects = [i['id'] for i in self.comm.backend.db.getView('viewDocType/x0')]
        status = exportELN(self.comm.backend, allProjects, fileName, docTypes)
        showMessage(self, 'Finished', status, 'Information')
    elif command[0] is Command.IMPORT:
      showMessage(self, 'Not fully implemented', 'The eln-file definition develops rapidly. We cannot adopt the import'+\
                        'in the same speed; hence it is disabled currently.', 'Warning')
      # fileName = QFileDialog.getOpenFileName(self, 'Load data from .eln file', str(Path.home()), '*.eln')[0]
      # if fileName != '':
      #   status = importELN(self.comm.backend, fileName)
      #   showMessage(self, 'Finished', status, 'Information')
      #   self.comm.changeSidebar.emit('redraw')
      #   self.comm.changeTable.emit('x0', '')
    elif command[0] is Command.EXIT:
      self.close()
    # view menu
    elif command[0] is Command.VIEW:
      self.comm.changeTable.emit(command[1], '')
    # system menu
    elif command[0] is Command.PROJECT_GROUP:
      dialogPG = ProjectGroup(self.comm)
      dialogPG.exec()
    elif command[0] is Command.CHANGE_PG:
      self.backend.configuration['defaultProjectGroup'] = command[1]
      with open(Path.home() / '.pastaELN.json', 'w', encoding='utf-8') as fConf:
        fConf.write(json.dumps(self.backend.configuration, indent=2))
      restart()
    elif command[0] is Command.SYNC:
      report = self.comm.backend.replicateDB(progressBar=self.sidebar.progress)
      showMessage(self, 'Report of syncronization', report, style='QLabel {min-width: 450px}')
    elif command[0] is Command.DATAHIERARCHY:
      dataHierarchyForm = DataHierarchyEditorDialog(self.comm.backend.db)
      dataHierarchyForm.instance.exec()
      restart()
    elif command[0] is Command.DATAVERSE_CONFIG:
      self.dataverseConfig = ConfigDialog()
      self.dataverseConfig.show()
    elif command[0] is Command.DATAVERSE_MAIN:
      self.dataverseMainDialog = MainDialog(self.comm.backend)
      self.dataverseMainDialog.show()
    elif command[0] is Command.TEST1:
      fileName = QFileDialog.getOpenFileName(self, 'Open file for extractor test', str(Path.home()), '*.*')[0]
      report = self.comm.backend.testExtractor(fileName, outputStyle='html')
      showMessage(self, 'Report of extractor test', report)
    elif command[0] is Command.TEST2:
      self.comm.testExtractor.emit()
    elif command[0] is Command.UPDATE:
      reportDict = updateExtractorList(self.backend.extractorPath)
      messageWindow = ScrollMessageBox('Extractor list updated', reportDict,
                                       style='QScrollArea{min-width:600 px; min-height:400px}')
      messageWindow.exec()
      restart()
    elif command[0] is Command.CONFIG:
      dialogC = Configuration(self.comm)
      dialogC.exec()
    # remainder
    elif command[0] is Command.WEBSITE:
      webbrowser.open('https://pasta-eln.github.io/pasta-eln/')
    elif command[0] is Command.VERIFY_DB:
      report = self.comm.backend.checkDB(outputStyle='html', minimal=True)
      showMessage(self, 'Report of database verification', report, style='QLabel {min-width: 800px}')
    elif command[0] is Command.SHORTCUTS:
      showMessage(self, 'Keyboard shortcuts', shortcuts)
    elif command[0] is Command.RESTART:
      restart()
    else:
      print("**ERROR gui menu unknown:", command)
    return


def mainGUI() -> tuple[Union[QCoreApplication, None], MainWindow]:
  """
    Main method and entry point for commands

  Returns:
    MainWindow: main window
  """
  # logging has to be started first
  log_path = Path.home() / 'pastaELN.log'
  #  old versions of basicConfig do not know "encoding='utf-8'"
  logging.basicConfig(filename=log_path, level=logging.INFO, format='%(asctime)s|%(levelname)s:%(message)s',
                      datefmt='%m-%d %H:%M:%S')
  for package in ['urllib3', 'requests', 'asyncio', 'PIL', 'matplotlib']:
    logging.getLogger(package).setLevel(logging.WARNING)
  logging.info('Start PASTA GUI')
  # remainder
  if not QApplication.instance():
    application = QApplication().instance()
  else:
    application = QApplication.instance()
  main_window = MainWindow()
  logging.getLogger().setLevel(getattr(logging, main_window.backend.configuration['GUI']['loggingLevel']))
  theme = main_window.backend.configuration['GUI']['theme']
  if theme != 'none':
    apply_stylesheet(application, theme=f'{theme}.xml')
  # qtawesome and matplot cannot coexist
  import qtawesome as qta
  if not isinstance(qta.icon('fa5s.times'), QIcon):
    logging.error('qtawesome: could not load. Likely matplotlib is included and can not coexist.')
    print('qtawesome: could not load. Likely matplotlib is included and can not coexist.')
  # end test coexistence
  logging.info('End PASTA GUI')
  return application, main_window


class Command(Enum):
  """ Commands used in this file """
  EXPORT    = 1
  IMPORT    = 2
  EXIT      = 3
  VIEW      = 4
  PROJECT_GROUP = 5
  CHANGE_PG = 6
  SYNC      = 7
  DATAHIERARCHY = 8
  TEST1     = 9
  TEST2     = 10
  UPDATE    = 11
  CONFIG    = 12
  WEBSITE   = 13
  VERIFY_DB = 14
  SHORTCUTS = 15
  RESTART   = 16
  EXPORT_ALL= 17
  DATAVERSE_CONFIG = 18
  DATAVERSE_MAIN = 19


def startMain() -> None:
  """
  Main function to start GUI. Extra function is required to allow starting in module fashion
  """
  app, window = mainGUI()
  window.show()
  if app:
    app.exec()


# called by python3 -m pasta_eln.gui
if __name__ == '__main__':
  startMain()



""" Represents a base model object. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: base_model.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
from collections.abc import Generator
from typing import Any

from pasta_eln.dataverse.incorrect_parameter_error import IncorrectParameterError


class BaseModel:
  """
  Represents a base model object.

  Explanation:
      This class serves as a base for other model classes and provides common attributes and methods.

  """

  def __init__(self,
               _id: str | None = None,
               _rev: str | None = None):
    """
    Initializes a base model object.

    Args:
        _id (str | None): The ID of the object. Defaults to None.
        _rev (str | None): The revision of the object. Defaults to None.

    Raises:
        IncorrectParameterError: If the _id or _rev parameter is not of type str.

    """
    super().__init__()
    if isinstance(_id, str | None):
      self._id: str | None = _id
    else:
      raise IncorrectParameterError(f"Expected string type for id but got {type(_id)}")
    if isinstance(_rev, str | None):
      self._rev: str | None = _rev
    else:
      raise IncorrectParameterError(f"Expected string type for rev but got {type(_rev)}")

  def __iter__(self) -> Generator[tuple[str, Any], None, None]:
    """
    Iterates over the attributes of the object and yields key-value pairs.

    Yields:
        tuple[str, Any]: A tuple containing the attribute name and its corresponding value.

    """
    for key in self.__dict__:
      if key in ['_id', '_rev']:
        yield key, getattr(self, key)
      else:
        yield key[1:], getattr(self, key)

  @property
  def id(self) -> str | None:
    """
    Returns the ID of the object.

    Returns:
        str | None: The ID of the object.

    """

    return self._id

  @id.setter
  def id(self, value: str | None) -> None:
    """
    Sets the ID of the object.

    Args:
        value (str | None): The ID value to be set.

    """
    self._id = value

  @id.deleter
  def id(self) -> None:
    """
    Deletes the ID of the object.

    """
    del self._id

  @property
  def rev(self) -> str | None:
    """
    Returns the revision of the object.

    Returns:
        str | None: The revision of the object.

    """
    return self._rev

  @rev.setter
  def rev(self, value: str | None) -> None:
    """
    Sets the revision of the object.

    Args:
        value (str | None): The revision value to be set.

    """
    self._rev = value

  @rev.deleter
  def rev(self) -> None:
    """
    Deletes the revision of the object.

    """
    del self._rev



""" DeleteColumnDelegate for the table views """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2023
#
#  Author: Jithu Murugan
#  Filename: delete_column_delegate.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

import logging
from typing import Union

from PySide6.QtCore import QAbstractItemModel, QEvent, QModelIndex, QPersistentModelIndex, QSize, Signal
from PySide6.QtGui import QPainter
from PySide6.QtWidgets import QApplication, QPushButton, QStyle, QStyleOption, QStyleOptionButton, QStyleOptionViewItem, \
  QStyledItemDelegate, QWidget

from .utility_functions import is_click_within_bounds


class DeleteColumnDelegate(QStyledItemDelegate):
  """
  Delegate for creating the delete icon for the delete column in the data hierarchy table views
  """
  delete_clicked_signal = Signal(
    int)  # Signal to inform the delete button click with the position in the table as the parameter

  def __init__(self) -> None:
    """
      Constructor
    """
    super().__init__()
    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")

  def paint(self,
            painter: QPainter,
            option: QStyleOption,
            index: Union[QModelIndex, QPersistentModelIndex]) -> None:
    """
    Draws the delete button within the cell represented by index
    Args:
      painter (QPainter): Painter instance for painting the button.
      option (QStyleOptionViewItem): Style option for the cell represented by index.
      index (Union[QModelIndex, QPersistentModelIndex]): Cell index.

    Returns: None

    """
    button = QPushButton()
    opt = QStyleOptionButton()
    opt.state = QStyle.StateFlag.State_Active | QStyle.StateFlag.State_Enabled  # type: ignore[attr-defined]
    opt.rect = option.rect  # type: ignore[attr-defined]
    opt.icon = QApplication.style().standardIcon(  # type: ignore[attr-defined]
      QStyle.StandardPixmap.SP_DialogDiscardButton)
    opt.iconSize = QSize(15, 15)  # type: ignore[attr-defined]
    QApplication.style().drawControl(QStyle.ControlElement.CE_PushButton, opt, painter, button)

  def createEditor(self,
                   parent: QWidget,
                   option: QStyleOptionViewItem,
                   index: Union[QModelIndex, QPersistentModelIndex]) -> QWidget:
    """
    Disable the editor for the delete column by simply returning None
    Args:
      parent (QWidget): Parent table view.
      option (QStyleOptionViewItem): Style option for the cell represented by index.
      index (Union[QModelIndex, QPersistentModelIndex]): Cell index.

    Returns: None

    """
    return None  # type: ignore[return-value]

  def editorEvent(self,
                  event: QEvent,
                  model: QAbstractItemModel,
                  option: QStyleOptionViewItem,
                  index: Union[QModelIndex, QPersistentModelIndex]) -> bool:
    """
    In case of click detected within the cell represented by index, the respective delete signal is emitted
    Args:
      event (QEvent): The editor event information.
      model (QAbstractItemModel): Model data representing the table view.
      option (QStyleOptionViewItem): QStyleOption for the table cell.
      index (Union[QModelIndex, QPersistentModelIndex]): Table cell index.

    Returns (bool): True if deleted otherwise False

    """
    if is_click_within_bounds(event, option):
      row = index.row()
      self.logger.info("Delete signal emitted for the position: {%s}", row)
      self.delete_clicked_signal.emit(row)
      return True
    return False



""" DataHierarchyEditorDialog which is extended from the Ui_DataHierarchyEditorDialogBase """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2023
#
#  Author: Jithu Murugan
#  Filename: data_hierarchy_editor_dialog.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

import copy
import logging
import sys
import webbrowser
from typing import Any

from PySide6 import QtWidgets
from PySide6.QtCore import QCoreApplication, QObject, Signal, Slot
from PySide6.QtWidgets import QApplication, QLineEdit, QMessageBox
from cloudant.document import Document

from .attachments_tableview_data_model import AttachmentsTableViewModel
from .constants import ATTACHMENT_TABLE_DELETE_COLUMN_INDEX, \
  ATTACHMENT_TABLE_REORDER_COLUMN_INDEX, DATA_HIERARCHY_HELP_PAGE_URL, METADATA_TABLE_DELETE_COLUMN_INDEX, \
  METADATA_TABLE_IRI_COLUMN_INDEX, METADATA_TABLE_REORDER_COLUMN_INDEX, METADATA_TABLE_REQUIRED_COLUMN_INDEX
from .create_type_dialog import CreateTypeDialog
from .data_hierarchy_editor_dialog_base import Ui_DataHierarchyEditorDialogBase
from .delete_column_delegate import DeleteColumnDelegate
from .document_null_exception import DocumentNullException
from .generic_exception import GenericException
from .iri_column_delegate import IriColumnDelegate
from .key_not_found_exception import \
  KeyNotFoundException
from .lookup_iri_action import LookupIriAction
from .mandatory_column_delegate import MandatoryColumnDelegate
from .metadata_tableview_data_model import MetadataTableViewModel
from .reorder_column_delegate import ReorderColumnDelegate
from .utility_functions import adapt_type, adjust_data_hierarchy_data_to_v4, can_delete_type, \
  check_data_hierarchy_types, \
  generate_empty_type, \
  get_missing_metadata_message, get_next_possible_structural_level_title, \
  get_types_for_display, show_message
from ...database import Database


class DataHierarchyEditorDialog(Ui_DataHierarchyEditorDialogBase, QObject):
  """ DataHierarchyEditorDialog class which is extended from the Ui_DataHierarchyEditorDialogBase
  and contains the UI elements and related logic"""
  type_changed_signal = Signal(str)

  def __new__(cls, *_: Any, **__: Any) -> Any:
    """
    Instantiates the DataHierarchyEditorDialog
    """
    return super(DataHierarchyEditorDialog, cls).__new__(cls)

  def __init__(self, database: Database) -> None:
    """
    Constructs the data hierarchy data editor

    Args:
      database (Database): Pasta ELN database instance

    Raises:
      DocumentNullException: Raised when passed in argument database has null data_hierarchy_document.
    """
    super().__init__()
    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")

    self.data_hierarchy_loaded: bool = False
    self.data_hierarchy_types: Any = {}
    self.selected_type_metadata: dict[str, list[dict[str, Any]]] | Any = {}

    # Set up the UI elements
    self.instance = QtWidgets.QDialog()
    super().setupUi(self.instance)

    # Gets the data_hierarchy data from db and adjust the data to the latest version
    if database is None:
      raise GenericException("Null database instance passed to the initializer", {})

    self.database: Database = database
    self.data_hierarchy_document: Document = self.database.db['-dataHierarchy-']
    if not self.data_hierarchy_document:
      raise DocumentNullException("Null data_hierarchy document in db instance", {})

    # Instantiates metadata & attachment table models along with the column delegates
    self.metadata_table_data_model = MetadataTableViewModel()
    self.attachments_table_data_model = AttachmentsTableViewModel()

    self.required_column_delegate_metadata_table = MandatoryColumnDelegate()
    self.delete_column_delegate_metadata_table = DeleteColumnDelegate()
    self.reorder_column_delegate_metadata_table = ReorderColumnDelegate()
    self.iri_column_delegate_metadata_table = IriColumnDelegate()
    self.delete_column_delegate_attach_table = DeleteColumnDelegate()
    self.reorder_column_delegate_attach_table = ReorderColumnDelegate()

    self.typeMetadataTableView.setItemDelegateForColumn(METADATA_TABLE_REQUIRED_COLUMN_INDEX,
                                                        self.required_column_delegate_metadata_table)
    self.typeMetadataTableView.setItemDelegateForColumn(METADATA_TABLE_DELETE_COLUMN_INDEX,
                                                        self.delete_column_delegate_metadata_table)
    self.typeMetadataTableView.setItemDelegateForColumn(METADATA_TABLE_REORDER_COLUMN_INDEX,
                                                        self.reorder_column_delegate_metadata_table)
    self.typeMetadataTableView.setItemDelegateForColumn(METADATA_TABLE_IRI_COLUMN_INDEX,
                                                        self.iri_column_delegate_metadata_table)
    self.typeMetadataTableView.setModel(self.metadata_table_data_model)

    for column_index, width in self.metadata_table_data_model.column_widths.items():
      self.typeMetadataTableView.setColumnWidth(column_index, width)
    # When resized, only stretch the query column of typeMetadataTableView
    self.typeMetadataTableView.horizontalHeader().setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeMode.Stretch)

    self.typeAttachmentsTableView.setItemDelegateForColumn(
      ATTACHMENT_TABLE_DELETE_COLUMN_INDEX,
      self.delete_column_delegate_attach_table)
    self.typeAttachmentsTableView.setItemDelegateForColumn(
      ATTACHMENT_TABLE_REORDER_COLUMN_INDEX,
      self.reorder_column_delegate_attach_table)
    self.typeAttachmentsTableView.setModel(self.attachments_table_data_model)

    for column_index, width in self.attachments_table_data_model.column_widths.items():
      self.typeAttachmentsTableView.setColumnWidth(column_index, width)
    # When resized, only stretch the type column of typeAttachmentsTableView
    self.typeAttachmentsTableView.horizontalHeader().setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeMode.Stretch)

    # Create the dialog for new type creation
    self.create_type_dialog = CreateTypeDialog(self.create_type_accepted_callback, self.create_type_rejected_callback)

    # Set up the slots for the UI items
    self.setup_slots()

    # Hide the attachment table and the add attachment button initially
    self.addAttachmentPushButton.hide()
    self.typeAttachmentsTableView.hide()

    self.load_data_hierarchy_data()

  def type_combo_box_changed(self,
                             new_type_selected: Any) -> None:
    """
    Combobox value changed callback for the selected type
    Args:
      new_type_selected (Any): Newly set value for the combobox.

    Returns: Nothing

    Raises:
      KeyNotFoundException: Raised when passed in argument @new_type_selected is not found in data_hierarchy_types

    """
    self.logger.info("New type selected in UI: {%s}", new_type_selected)
    self.clear_ui()
    new_type_selected = adapt_type(new_type_selected)
    self.type_changed_signal.emit(new_type_selected)
    if new_type_selected and self.data_hierarchy_types:
      if new_type_selected not in self.data_hierarchy_types:
        raise KeyNotFoundException(f"Key {new_type_selected} "
                                   f"not found in data_hierarchy_types", {})
      selected_type = self.data_hierarchy_types.get(new_type_selected)
      # Get the metadata for the selected type and store the list in selected_type_metadata
      self.selected_type_metadata = selected_type.get("meta")

      # Type displayed_title is set in a line edit
      self.typeDisplayedTitleLineEdit.setText(selected_type.get('title'))

      # Type IRI is set in a line edit
      self.typeIriLineEdit.setText(selected_type.get('IRI'))

      # Gets the attachment data from selected type and set it in table view
      self.attachments_table_data_model.update(selected_type.get('attachments'))

      # Reset the metadata group combo-box
      self.metadataGroupComboBox.addItems(list(self.selected_type_metadata.keys())
                                          if self.selected_type_metadata else [])
      self.metadataGroupComboBox.setCurrentIndex(0)

  def set_iri_lookup_action(self,
                            lookup_term: str) -> None:
    """
    Sets the IRI lookup action for the IRI line edit
    Args:
      lookup_term (str): Default lookup term to be used by the lookup service

    Returns: Nothing

    """
    for act in self.typeIriLineEdit.actions():
      if isinstance(act, LookupIriAction):
        act.deleteLater()
    self.typeIriLineEdit.addAction(
      LookupIriAction(parent_line_edit=self.typeIriLineEdit, lookup_term=lookup_term),
      QLineEdit.ActionPosition.TrailingPosition)

  def metadata_group_combo_box_changed(self,
                                       new_selected_metadata_group: Any) -> None:
    """
    Combobox value changed callback for the selected type metadata groups
    Args:
      new_selected_metadata_group (Any): Newly set value for the combobox.

    Returns: Nothing
    """
    self.logger.info("New metadata group selected in UI: {%s}", new_selected_metadata_group)
    if new_selected_metadata_group and self.selected_type_metadata:
      # Update the metadata table as per the selected metadata group from combobox
      self.metadata_table_data_model.update(self.selected_type_metadata.get(new_selected_metadata_group))

  def add_new_metadata_group(self) -> None:
    """
    Click event handler for adding new metadata group
    Returns: Nothing
    """
    new_group = self.addMetadataGroupLineEdit.text()
    if not new_group:
      show_message("Enter non-null/valid metadata group name!!.....", QMessageBox.Icon.Warning)
      return None
    if not self.data_hierarchy_loaded or self.data_hierarchy_types is None:
      show_message("Load the data hierarchy data first....", QMessageBox.Icon.Warning)
      return None
    if new_group in self.selected_type_metadata.keys():
      show_message("Metadata group already exists....", QMessageBox.Icon.Warning)
      return None
    # Add the new group to the metadata list and refresh the group combo box
    self.logger.info("User added new metadata group: {%s}", new_group)
    self.selected_type_metadata[new_group] = []
    self.metadataGroupComboBox.clear()
    self.metadataGroupComboBox.addItems(list(self.selected_type_metadata.keys()))
    self.metadataGroupComboBox.setCurrentIndex(len(self.selected_type_metadata.keys()) - 1)
    return None

  def delete_selected_metadata_group(self) -> None:
    """
    Click event handler for deleting the selected metadata group
    Returns: Nothing
    """
    selected_group = self.metadataGroupComboBox.currentText()
    if self.selected_type_metadata is None:
      show_message("Load the data hierarchy data first....", QMessageBox.Icon.Warning)
      return None
    if selected_group and selected_group in self.selected_type_metadata.keys():
      self.logger.info("User deleted the selected metadata group: {%s}", selected_group)
      self.selected_type_metadata.pop(selected_group)
      self.metadataGroupComboBox.clear()
      self.typeMetadataTableView.model().update([])
      self.metadataGroupComboBox.addItems(list(self.selected_type_metadata.keys()))
      self.metadataGroupComboBox.setCurrentIndex(len(self.selected_type_metadata.keys()) - 1)
    return None

  def update_type_displayed_title(self,
                                  modified_type_displayed_title: str) -> None:
    """
    Value changed callback for the type displayed title line edit

    Args:
        modified_type_displayed_title (str): Modified data_hierarchy type displayed title

    Returns: Nothing
    """
    current_type = self.typeComboBox.currentText()
    current_type = adapt_type(current_type)
    if modified_type_displayed_title is not None and current_type in self.data_hierarchy_types:
      self.data_hierarchy_types.get(current_type)["title"] = modified_type_displayed_title
      self.set_iri_lookup_action(modified_type_displayed_title)

  def update_type_iri(self,
                      modified_iri: str) -> None:
    """
    Value changed callback for the IRI line edit

    Args:
        modified_iri (str): Modified IRI to be set for the selected type

    Returns: Nothing
    """
    current_type = self.typeComboBox.currentText()
    current_type = adapt_type(current_type)
    if modified_iri is not None and current_type in self.data_hierarchy_types:
      self.data_hierarchy_types.get(current_type)["IRI"] = modified_iri

  def delete_selected_type(self) -> None:
    """
    Delete the selected type from the type selection combo-box and also from the loaded data_hierarchy_types

    Returns: Nothing
    """
    selected_type = self.typeComboBox.currentText()
    selected_type = adapt_type(selected_type)
    if not self.data_hierarchy_loaded:
      show_message("Load the data hierarchy data first....", QMessageBox.Icon.Warning)
      return
    if self.data_hierarchy_types is None or self.data_hierarchy_document is None:
      show_message("Load the data hierarchy data first....", QMessageBox.Icon.Warning)
      return
    if selected_type and selected_type in self.data_hierarchy_types:
      self.logger.info("User deleted the selected type: {%s}", selected_type)
      self.data_hierarchy_types.pop(selected_type)
      self.typeComboBox.clear()
      self.typeComboBox.addItems(get_types_for_display(self.data_hierarchy_types.keys()))
      self.typeComboBox.setCurrentIndex(0)

  def clear_ui(self) -> None:
    """
    Clear the UI elements including the tables.
    Invoked when the type combobox selection changes
    Returns: None

    """
    # Disable the signals for the line edits before clearing in order to avoid clearing the respective
    # iri and displayed_titles for the selected type from data_hierarchy document
    self.typeDisplayedTitleLineEdit.textChanged[str].disconnect()
    self.typeIriLineEdit.textChanged[str].disconnect()
    self.typeDisplayedTitleLineEdit.clear()
    self.typeIriLineEdit.clear()
    self.typeDisplayedTitleLineEdit.textChanged[str].connect(self.update_type_displayed_title)
    self.typeIriLineEdit.textChanged[str].connect(self.update_type_iri)

    self.metadataGroupComboBox.clear()
    self.addMetadataGroupLineEdit.clear()
    self.typeMetadataTableView.model().update([])
    self.typeAttachmentsTableView.model().update([])

  def create_type_accepted_callback(self) -> None:
    """
    Callback for the OK button of CreateTypeDialog to create a new type in the data_hierarchy data set

    Returns: Nothing
    """
    title = self.create_type_dialog.next_struct_level \
      if self.create_type_dialog.structuralLevelCheckBox.isChecked() \
      else self.create_type_dialog.titleLineEdit.text()
    displayed_title = self.create_type_dialog.displayedTitleLineEdit.text()
    self.create_type_dialog.clear_ui()
    self.create_new_type(title, displayed_title)

  def create_type_rejected_callback(self) -> None:
    """
    Callback for the cancel button of CreateTypeDialog

    Returns: Nothing
    """
    self.create_type_dialog.clear_ui()

  def show_create_type_dialog(self) -> None:
    """
    Opens a dialog which allows the user to enter the details to create a new type (structural or normal)
    Returns: Nothing
    """
    if self.data_hierarchy_types is not None and self.data_hierarchy_loaded:
      structural_title = get_next_possible_structural_level_title(self.data_hierarchy_types.keys())
      self.create_type_dialog.set_structural_level_title(structural_title)
      self.create_type_dialog.show()
    else:
      show_message("Load the data hierarchy data first...", QMessageBox.Icon.Warning)

  def setup_slots(self) -> None:
    """
    Set up the slots for the UI elements of data hierarchy editor
    Returns: Nothing
    """
    self.logger.info("Setting up slots for the editor..")
    # Slots for the buttons
    self.addMetadataRowPushButton.clicked.connect(self.metadata_table_data_model.add_data_row)
    self.addAttachmentPushButton.clicked.connect(self.attachments_table_data_model.add_data_row)
    self.saveDataHierarchyPushButton.clicked.connect(self.save_data_hierarchy)
    self.addMetadataGroupPushButton.clicked.connect(self.add_new_metadata_group)
    self.deleteMetadataGroupPushButton.clicked.connect(self.delete_selected_metadata_group)
    self.deleteTypePushButton.clicked.connect(self.delete_selected_type)
    self.addTypePushButton.clicked.connect(self.show_create_type_dialog)
    self.cancelPushButton.clicked.connect(self.instance.close)
    self.helpPushButton.clicked.connect(lambda: webbrowser.open(DATA_HIERARCHY_HELP_PAGE_URL))
    self.attachmentsShowHidePushButton.clicked.connect(self.show_hide_attachments_table)

    # Slots for the combo-boxes
    self.typeComboBox.currentTextChanged.connect(self.type_combo_box_changed)
    self.metadataGroupComboBox.currentTextChanged.connect(self.metadata_group_combo_box_changed)

    # Slots for line edits
    self.typeDisplayedTitleLineEdit.textChanged[str].connect(self.update_type_displayed_title)
    self.typeIriLineEdit.textChanged[str].connect(self.update_type_iri)

    # Slots for the delegates
    self.delete_column_delegate_metadata_table.delete_clicked_signal.connect(self.metadata_table_data_model.delete_data)
    self.reorder_column_delegate_metadata_table.re_order_signal.connect(self.metadata_table_data_model.re_order_data)

    self.delete_column_delegate_attach_table.delete_clicked_signal.connect(
      self.attachments_table_data_model.delete_data)
    self.reorder_column_delegate_attach_table.re_order_signal.connect(self.attachments_table_data_model.re_order_data)

    self.type_changed_signal.connect(self.check_and_disable_delete_button)

  def load_data_hierarchy_data(self) -> None:
    """
    Load button click event handler which loads the data in the UI
    Returns:

    """
    self.logger.info("User loaded the data hierarchy data in UI")
    if self.data_hierarchy_document is None:
      raise GenericException("Null data_hierarchy_document, erroneous app state", {})
    # Load the data_hierarchy types from the db document
    for data in self.data_hierarchy_document:
      if isinstance(self.data_hierarchy_document[data], dict):
        self.data_hierarchy_types[data] = copy.deepcopy(self.data_hierarchy_document[data])
    adjust_data_hierarchy_data_to_v4(self.data_hierarchy_types)
    self.data_hierarchy_loaded = True

    # Set the types in the type selector combo-box
    self.typeComboBox.clear()
    self.typeComboBox.addItems(get_types_for_display(self.data_hierarchy_types.keys()))
    self.typeComboBox.setCurrentIndex(0)

  def save_data_hierarchy(self) -> None:
    """
    Save the modified data hierarchy document data in database
    """
    self.logger.info("User clicked the save button..")
    types_with_missing_metadata, types_with_null_name_metadata, types_with_duplicate_metadata \
      = check_data_hierarchy_types(self.data_hierarchy_types)
    if (types_with_missing_metadata
        or types_with_null_name_metadata
        or types_with_duplicate_metadata):
      message = get_missing_metadata_message(types_with_missing_metadata,
                                             types_with_null_name_metadata,
                                             types_with_duplicate_metadata)
      show_message(message, QMessageBox.Icon.Warning)
      self.logger.warning(message)
      return

    result = show_message("Save will close the tool and restart the Pasta Application (Yes/No?)",
                          QMessageBox.Icon.Question,
                          QMessageBox.StandardButton.No | QMessageBox.StandardButton.Yes,
                          QMessageBox.StandardButton.Yes)

    if result == QMessageBox.StandardButton.Yes:
      # Clear all the data from the data_hierarchy_document
      for data in list(self.data_hierarchy_document.keys()):
        if isinstance(self.data_hierarchy_document[data], dict):
          del self.data_hierarchy_document[data]
      # Copy all the modifications
      for type_name, type_structure in self.data_hierarchy_types.items():
        self.data_hierarchy_document[type_name] = type_structure
      # Save the modified document
      self.data_hierarchy_document.save()
      self.database.dataHierarchy = dict(self.data_hierarchy_document)
      self.database.initDocTypeViews(16)
      self.instance.close()

  def create_new_type(self,
                      title: str,
                      displayed_title: str) -> None:
    """
    Add a new type to the loaded data_hierarchy_data from the db
    Args:
      title (str): The new key entry used for the data_hierarchy_data
      displayed_title (str): The new displayed_title set for the new type entry in data_hierarchy_data

    Returns:

    """
    if self.data_hierarchy_document is None or self.data_hierarchy_types is None:
      self.logger.error("Null data_hierarchy_document/data_hierarchy_types, erroneous app state")
      raise GenericException("Null data_hierarchy_document/data_hierarchy_types, erroneous app state", {})
    if title in self.data_hierarchy_types:
      show_message(
        f"Type (title: {title} displayed title: {displayed_title}) cannot be added since it exists in DB already....",
        QMessageBox.Icon.Warning)
    else:
      if not title:
        self.logger.warning("Enter non-null/valid title!!.....")
        show_message("Enter non-null/valid title!!.....", QMessageBox.Icon.Warning)
        return
      self.logger.info("User created a new type and added "
                       "to the data_hierarchy document: Title: {%s}, Displayed Title: {%s}", title, displayed_title)
      empty_type = generate_empty_type(displayed_title)
      self.data_hierarchy_types[title] = empty_type
      self.typeComboBox.clear()
      self.typeComboBox.addItems(get_types_for_display(self.data_hierarchy_types.keys()))
      self.typeComboBox.setCurrentIndex(len(self.data_hierarchy_types) - 1)

  def show_hide_attachments_table(self) -> None:
    """
    Show/hide the attachment table and the add attachment button
    Returns: Nothing
    """
    self.typeAttachmentsTableView.setVisible(not self.typeAttachmentsTableView.isVisible())
    self.addAttachmentPushButton.setVisible(not self.addAttachmentPushButton.isVisible())

  @Slot(str)
  def check_and_disable_delete_button(self,
                                      selected_type: str) -> None:
    """
    Slot to check and disable the "delete" button for the selected type
    Args:
      selected_type (str): Selected type: typesComboBox

    Returns: Nothing

    """
    (self.deleteTypePushButton
     .setEnabled(can_delete_type(self.data_hierarchy_types.keys(),
                                 selected_type)))


def get_gui(database: Database) -> tuple[
  QCoreApplication | QApplication, QtWidgets.QDialog, DataHierarchyEditorDialog]:
  """
  Creates the editor UI and return it
  Args:
    database (Database): PASTA ELN Database instance.
  Returns:

  """
  instance = QApplication.instance()
  application = QApplication(sys.argv) if instance is None else instance
  data_hierarchy_form: DataHierarchyEditorDialog = DataHierarchyEditorDialog(database)

  return application, data_hierarchy_form.instance, data_hierarchy_form



# Form implementation generated from reading ui file 'upload_config_dialog_base.ui'
#
# Created by: PyQt6 UI code generator 6.4.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PySide6 import QtCore, QtGui, QtWidgets


class Ui_UploadConfigDialog(object):
  def setupUi(self, UploadConfigDialog):
    UploadConfigDialog.setObjectName("UploadConfigDialog")
    UploadConfigDialog.resize(786, 226)
    self.verticalLayout = QtWidgets.QVBoxLayout(UploadConfigDialog)
    self.verticalLayout.setObjectName("verticalLayout")
    self.mainVerticalLayout = QtWidgets.QVBoxLayout()
    self.mainVerticalLayout.setObjectName("mainVerticalLayout")
    self.parallelUploadsHorizontalLayout = QtWidgets.QHBoxLayout()
    self.parallelUploadsHorizontalLayout.setObjectName("parallelUploadsHorizontalLayout")
    self.numParallelLabel = QtWidgets.QLabel(parent=UploadConfigDialog)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.numParallelLabel.sizePolicy().hasHeightForWidth())
    self.numParallelLabel.setSizePolicy(sizePolicy)
    self.numParallelLabel.setObjectName("numParallelLabel")
    self.parallelUploadsHorizontalLayout.addWidget(self.numParallelLabel)
    self.numParallelComboBox = QtWidgets.QComboBox(parent=UploadConfigDialog)
    self.numParallelComboBox.setObjectName("numParallelComboBox")
    self.parallelUploadsHorizontalLayout.addWidget(self.numParallelComboBox)
    self.mainVerticalLayout.addLayout(self.parallelUploadsHorizontalLayout)
    self.projectContentsScrollArea = QtWidgets.QScrollArea(parent=UploadConfigDialog)
    self.projectContentsScrollArea.setWidgetResizable(True)
    self.projectContentsScrollArea.setObjectName("projectContentsScrollArea")
    self.scrollAreaWidgetContents = QtWidgets.QWidget()
    self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 750, 159))
    self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
    self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
    self.verticalLayout_3.setObjectName("verticalLayout_3")
    self.projectItemsVerticalLayout = QtWidgets.QVBoxLayout()
    self.projectItemsVerticalLayout.setObjectName("projectItemsVerticalLayout")
    self.measurementCheckBox = QtWidgets.QCheckBox(parent=self.scrollAreaWidgetContents)
    self.measurementCheckBox.setChecked(True)
    self.measurementCheckBox.setObjectName("measurementCheckBox")
    self.projectItemsVerticalLayout.addWidget(self.measurementCheckBox)
    self.samplesCheckBox = QtWidgets.QCheckBox(parent=self.scrollAreaWidgetContents)
    self.samplesCheckBox.setChecked(True)
    self.samplesCheckBox.setObjectName("samplesCheckBox")
    self.projectItemsVerticalLayout.addWidget(self.samplesCheckBox)
    self.proceduresCheckBox = QtWidgets.QCheckBox(parent=self.scrollAreaWidgetContents)
    self.proceduresCheckBox.setChecked(True)
    self.proceduresCheckBox.setObjectName("proceduresCheckBox")
    self.projectItemsVerticalLayout.addWidget(self.proceduresCheckBox)
    self.instrumentsCheckBox = QtWidgets.QCheckBox(parent=self.scrollAreaWidgetContents)
    self.instrumentsCheckBox.setChecked(True)
    self.instrumentsCheckBox.setObjectName("instrumentsCheckBox")
    self.projectItemsVerticalLayout.addWidget(self.instrumentsCheckBox)
    self.unidentifiedCheckBox = QtWidgets.QCheckBox(parent=self.scrollAreaWidgetContents)
    self.unidentifiedCheckBox.setChecked(True)
    self.unidentifiedCheckBox.setObjectName("unidentifiedCheckBox")
    self.projectItemsVerticalLayout.addWidget(self.unidentifiedCheckBox)
    self.verticalLayout_3.addLayout(self.projectItemsVerticalLayout)
    self.projectContentsScrollArea.setWidget(self.scrollAreaWidgetContents)
    self.mainVerticalLayout.addWidget(self.projectContentsScrollArea)
    self.verticalLayout.addLayout(self.mainVerticalLayout)
    self.buttonBox = QtWidgets.QDialogButtonBox(parent=UploadConfigDialog)
    self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
    self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Save)
    self.buttonBox.setObjectName("buttonBox")
    self.verticalLayout.addWidget(self.buttonBox)

    self.retranslateUi(UploadConfigDialog)
    self.buttonBox.accepted.connect(UploadConfigDialog.accept) # type: ignore
    self.buttonBox.rejected.connect(UploadConfigDialog.reject) # type: ignore
    QtCore.QMetaObject.connectSlotsByName(UploadConfigDialog)

  def retranslateUi(self, UploadConfigDialog):
    _translate = QtCore.QCoreApplication.translate
    UploadConfigDialog.setWindowTitle(_translate("UploadConfigDialog", "Configure project upload"))
    UploadConfigDialog.setToolTip(_translate("UploadConfigDialog", "Select the configuration parameters used for dataverse upload."))
    self.numParallelLabel.setText(_translate("UploadConfigDialog", "No of parallel uploads"))
    self.numParallelComboBox.setToolTip(_translate("UploadConfigDialog", "Choose the number of parallel dataverse uploads."))
    self.projectContentsScrollArea.setToolTip(_translate("UploadConfigDialog", "Select the sub-items in a project to be uploaded to dataverse."))
    self.measurementCheckBox.setText(_translate("UploadConfigDialog", "Measurements"))
    self.samplesCheckBox.setText(_translate("UploadConfigDialog", "Samples"))
    self.proceduresCheckBox.setText(_translate("UploadConfigDialog", "Procedures"))
    self.instrumentsCheckBox.setText(_translate("UploadConfigDialog", "Instruments"))
    self.unidentifiedCheckBox.setText(_translate("UploadConfigDialog", "Unidentified"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    UploadConfigDialog = QtWidgets.QDialog()
    ui = Ui_UploadConfigDialog()
    ui.setupUi(UploadConfigDialog)
    UploadConfigDialog.show()
    sys.exit(app.exec())



# -*- coding: utf-8 -*-
#
# PASTA-ELN documentation build configuration file, created by
# sphinx-quickstart on Tue Oct 13 08:41:19 2015.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import datetime
import sys
from os import pardir
from os.path import (
    abspath,
    dirname,
    exists,
    join as opj,
)

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#sys.path.insert(0, os.path.abspath('.'))

# generate missing pieces
for setup_py_path in (opj(pardir, 'setup.py'),  # travis
                      opj(pardir, pardir, 'setup.py')):  # RTD
    if exists(setup_py_path):
        sys.path.insert(0, abspath(dirname(setup_py_path)))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    #'sphinx.ext.autosummary',
    'sphinx.ext.autosectionlabel',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.ifconfig',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
]

# for the module reference
autosummary_generate = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'PASTA-ELN'
copyright = u'2022-{}, PASTA-ELN team'.format(datetime.datetime.now().year)
author = u'PASTA-ELN team'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
version = "2.6.0b1"
release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'<name>': ('https://docs.python.org/', None)}

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_rtd_theme'

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = '_static/pasta_logo.svg'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If true, the index is split into individual pages for each letter.
html_split_index = True

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# smart quotes are incompatible with the RST flavor of the generated manpages
# but see `smartquotes_action` for more fine-grained control, in case
# some of this functionality is needed
smartquotes = False



"""extract data from vendor
- Markdown .md for procedures
"""

def use(filePath, recipe='', saveFileName=None):
  """
  Args:
    filePath (Path): full path file name
    recipe (string): supplied to guide recipes
                     recipe is / separated hierarchical elements parent->child
    saveFileName (string): if given, save the image to this file-name
  Returns:
    dict: containing image, metaVendor, metaUser, recipe
  """
  with open(filePath,'r', encoding='utf-8') as fIn:
    return {'content':fIn.read(), 'recipe': 'procedure/markdown', 'metaVendor':{}, 'metaUser':{}}

  #other datatypes follow here
  #...

  #final return if nothing successful
  # return {'recipe': '-', 'metaVendor':{}, 'metaUser':{}}
